"""start

Revision ID: 0001
Revises: 
Create Date: 2016-04-13 19:54:05.937035

"""

# revision identifiers, used by Alembic.
revision = '0001'
down_revision = None
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import whisker


def upgrade():
    ### commands auto generated by Alembic - please adjust! ###
    op.create_table('master_config',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('name', sa.String(length=255), nullable=True),
    sa.Column('modified_at', whisker.sqlalchemy.ArrowMicrosecondType(), nullable=True),
    sa.Column('server', sa.String(length=255), nullable=True),
    sa.Column('port', sa.Integer(), nullable=True),
    sa.Column('wcm_prefix', sa.String(length=255), nullable=True),
    sa.Column('rfid_effective_time_s', sa.Float(), nullable=True),
    sa.PrimaryKeyConstraint('id', name=op.f('pk_master_config'))
    )
    op.create_table('rfidreader_config',
    sa.Column('port', sa.String(length=255), nullable=True),
    sa.Column('baudrate', sa.Integer(), nullable=True),
    sa.Column('bytesize', sa.Integer(), nullable=True),
    sa.Column('parity', sa.String(length=1), nullable=True),
    sa.Column('stopbits', sa.Float(), nullable=True),
    sa.Column('xonxoff', sa.Boolean(), nullable=True),
    sa.Column('rtscts', sa.Boolean(), nullable=True),
    sa.Column('dsrdtr', sa.Boolean(), nullable=True),
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('master_config_id', sa.Integer(), nullable=True),
    sa.Column('name', sa.String(length=255), nullable=True),
    sa.Column('enabled', sa.Boolean(), nullable=True),
    sa.ForeignKeyConstraint(['master_config_id'], ['master_config.id'], name=op.f('fk_rfidreader_config_master_config_id_master_config')),
    sa.PrimaryKeyConstraint('id', name=op.f('pk_rfidreader_config'))
    )
    op.create_table('balance_config',
    sa.Column('port', sa.String(length=255), nullable=True),
    sa.Column('baudrate', sa.Integer(), nullable=True),
    sa.Column('bytesize', sa.Integer(), nullable=True),
    sa.Column('parity', sa.String(length=1), nullable=True),
    sa.Column('stopbits', sa.Float(), nullable=True),
    sa.Column('xonxoff', sa.Boolean(), nullable=True),
    sa.Column('rtscts', sa.Boolean(), nullable=True),
    sa.Column('dsrdtr', sa.Boolean(), nullable=True),
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('master_config_id', sa.Integer(), nullable=True),
    sa.Column('reader_id', sa.Integer(), nullable=True),
    sa.Column('name', sa.String(length=255), nullable=True),
    sa.Column('enabled', sa.Boolean(), nullable=True),
    sa.Column('measurement_rate_hz', sa.Integer(), nullable=True),
    sa.Column('stability_n', sa.Integer(), nullable=True),
    sa.Column('tolerance_kg', sa.Float(), nullable=True),
    sa.Column('min_mass_kg', sa.Float(), nullable=True),
    sa.Column('unlock_mass_kg', sa.Float(), nullable=True),
    sa.Column('refload_mass_kg', sa.Float(), nullable=True),
    sa.Column('zero_value', sa.Integer(), nullable=True),
    sa.Column('refload_value', sa.Integer(), nullable=True),
    sa.Column('read_continuously', sa.Boolean(), nullable=True),
    sa.Column('amp_signal_filter_mode', sa.Integer(), nullable=True),
    sa.Column('fast_response_filter', sa.Boolean(), nullable=True),
    sa.ForeignKeyConstraint(['master_config_id'], ['master_config.id'], name=op.f('fk_balance_config_master_config_id_master_config')),
    sa.ForeignKeyConstraint(['reader_id'], ['rfidreader_config.id'], name=op.f('fk_balance_config_reader_id_rfidreader_config')),
    sa.PrimaryKeyConstraint('id', name=op.f('pk_balance_config'))
    )
    op.create_table('rfid_event',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('reader_id', sa.Integer(), nullable=True),
    sa.Column('rfid', sa.BigInteger(), nullable=True),
    sa.Column('first_detected_at', whisker.sqlalchemy.ArrowMicrosecondType(), nullable=True),
    sa.Column('last_detected_at', whisker.sqlalchemy.ArrowMicrosecondType(), nullable=True),
    sa.Column('n_events', sa.Integer(), nullable=True),
    sa.ForeignKeyConstraint(['reader_id'], ['rfidreader_config.id'], name=op.f('fk_rfid_event_reader_id_rfidreader_config')),
    sa.PrimaryKeyConstraint('id', name=op.f('pk_rfid_event'))
    )
    op.create_table('mass_event',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('rfid', sa.BigInteger(), nullable=True),
    sa.Column('reader_id', sa.Integer(), nullable=True),
    sa.Column('balance_id', sa.Integer(), nullable=True),
    sa.Column('at', whisker.sqlalchemy.ArrowMicrosecondType(), nullable=True),
    sa.Column('mass_kg', sa.Float(), nullable=True),
    sa.ForeignKeyConstraint(['balance_id'], ['balance_config.id'], name=op.f('fk_mass_event_balance_id_balance_config')),
    sa.ForeignKeyConstraint(['reader_id'], ['rfidreader_config.id'], name=op.f('fk_mass_event_reader_id_rfidreader_config')),
    sa.PrimaryKeyConstraint('id', name=op.f('pk_mass_event'))
    )
    ### end Alembic commands ###


def downgrade():
    ### commands auto generated by Alembic - please adjust! ###
    op.drop_table('mass_event')
    op.drop_table('rfid_event')
    op.drop_table('balance_config')
    op.drop_table('rfidreader_config')
    op.drop_table('master_config')
    ### end Alembic commands ###



#!/usr/bin/env python
# starfeeder/models.py

"""
    Copyright (C) 2015-2015 Rudolf Cardinal (rudolf@pobox.com).

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

import datetime
import logging
import serial
from typing import Any, Dict, List, Optional

import arrow
from sqlalchemy import (
    BigInteger,
    Boolean,
    Column,
    Float,
    ForeignKey,
    Integer,
    MetaData,
    String,
)
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import backref, relationship
from sqlalchemy.orm.session import Session
from whisker.sqlalchemy import (
    ALEMBIC_NAMING_CONVENTION,
    SqlAlchemyAttrDictMixin,
    ArrowMicrosecondType,
)
from whisker.lang import (
    ordered_repr,
    simple_repr,
    trunc_if_integer,
)

from starfeeder.constants import DEFAULT_BALANCE_READ_FREQUENCY_HZ

log = logging.getLogger(__name__)


# =============================================================================
# Constants
# =============================================================================

MAX_GENERIC_STRING_LENGTH = 255

# =============================================================================
# SQLAlchemy base.
# =============================================================================
# Derived classes will share the specified metadata.

MASTER_META = MetaData(naming_convention=ALEMBIC_NAMING_CONVENTION)
Base = declarative_base(metadata=MASTER_META)


# =============================================================================
# Mixin: serial port config
# =============================================================================
# http://docs.sqlalchemy.org/en/latest/orm/extensions/declarative/mixins.html

class SerialPortConfigMixin(object):
    # Note that defaults apply at insertion. So a class constructed within
    # Python doesn't have defaults set (until you save AND RELOAD somehow),
    # so we are better off using the __init__ function for defaults.
    # The __init__ function is free-form:
    # http://docs.sqlalchemy.org/en/rel_0_8/orm/mapper_config.html#constructors-and-object-initialization  # noqa
    port = Column(String(MAX_GENERIC_STRING_LENGTH))
    baudrate = Column(Integer)
    bytesize = Column(Integer)
    parity = Column(String(length=1))
    stopbits = Column(Float)
    xonxoff = Column(Boolean)
    rtscts = Column(Boolean)
    dsrdtr = Column(Boolean)

    def __init__(self,
                 port: str = '',
                 baudrate: int = 9600,
                 bytesize: int = serial.EIGHTBITS,
                 parity: str = serial.PARITY_NONE,
                 stopbits: float = serial.STOPBITS_ONE,
                 xonxoff: bool = False,
                 rtscts: bool = True,
                 dsrdtr: bool = False,
                 **kwargs) -> None:
        super().__init__(**kwargs)
        self.port = port
        self.baudrate = baudrate
        self.bytesize = bytesize
        self.parity = parity
        self.stopbits = stopbits
        self.xonxoff = xonxoff
        self.rtscts = rtscts
        self.dsrdtr = dsrdtr

    @staticmethod
    def fields_component_serial_port() -> List[str]:
        return ['port', 'bytesize', 'parity', 'stopbits',
                'xonxoff', 'rtscts', 'dsrdtr']

    def str_component_serial_port(self) -> str:
        flowmethods = []
        if self.xonxoff:
            flowmethods.append("XON/OFF")
        if self.rtscts:
            flowmethods.append("RTS/CTS")
        if self.dsrdtr:
            flowmethods.append("DTR/DSR")
        if flowmethods:
            flow = "+".join(flowmethods)
        else:
            flow = "no flow control"
        return "{port}, {speed} {bits}{parity}{stop}, {flow}".format(
            port=self.port,
            speed=self.baudrate,
            bits=self.bytesize,
            parity=self.parity,
            stop=trunc_if_integer(self.stopbits),
            flow=flow,
        )

    def get_serial_args(self) -> Dict[str, Any]:
        return dict(
            port=self.port,
            baudrate=self.baudrate,
            bytesize=self.bytesize,
            parity=self.parity,
            stopbits=self.stopbits,
            xonxoff=self.xonxoff,
            rtscts=self.rtscts,
            dsrdtr=self.dsrdtr,
        )


# =============================================================================
# RFID reader
# =============================================================================

class RfidReaderConfig(SqlAlchemyAttrDictMixin, SerialPortConfigMixin, Base):
    __tablename__ = 'rfidreader_config'
    id = Column(Integer, primary_key=True)
    master_config_id = Column(Integer, ForeignKey('master_config.id'))
    name = Column(String(MAX_GENERIC_STRING_LENGTH))
    enabled = Column(Boolean)

    def __init__(self, **kwargs) -> None:
        self.name = kwargs.pop('name', '')
        self.master_config_id = kwargs.pop('master_config_id')
        self.enabled = kwargs.pop('enabled', True)
        kwargs.setdefault('baudrate', 9600)
        kwargs.setdefault('bytesize', serial.EIGHTBITS)
        kwargs.setdefault('parity', serial.PARITY_NONE)
        kwargs.setdefault('stopbits', serial.STOPBITS_ONE)
        kwargs.setdefault('xonxoff', False)
        kwargs.setdefault('rtscts', False)  # works under Linux, not Windows
        kwargs.setdefault('dsrdtr', False)
        super().__init__(**kwargs)

    def __repr__(self) -> str:
        return ordered_repr(
            self,
            ['id', 'master_config_id', 'name', 'enabled'] +
            self.fields_component_serial_port())

    def __str__(self) -> str:
        return "{}{}: {}".format(
            "[DISABLED] " if not self.enabled else "",
            self.name or "[no name]",
            self.str_component_serial_port())

    # noinspection PyShadowingBuiltins
    @classmethod
    def get_name_from_id(cls, session: Session, id: int) -> Optional[str]:
        obj = session.query(cls).get(id)
        if obj is None:
            return None
        return obj.name


# =============================================================================
# Balance
# =============================================================================

class CalibrationReport(object):
    def __init__(self,
                 balance_id: int,
                 balance_name: str,
                 zero_value: int,
                 refload_value: int,
                 refload_mass_kg: float) -> None:
        self.balance_id = balance_id
        self.balance_name = balance_name
        self.zero_value = zero_value
        self.refload_value = refload_value
        self.refload_mass_kg = refload_mass_kg

    def __str__(self) -> str:
        return (
            "Calibrated balance {}: zero value = {}, "
            "loaded value = {} (at {} kg)".format(
                self.balance_name, self.zero_value,
                self.refload_value, self.refload_mass_kg))


class BalanceConfig(SqlAlchemyAttrDictMixin, SerialPortConfigMixin, Base):
    __tablename__ = 'balance_config'
    id = Column(Integer, primary_key=True)
    master_config_id = Column(Integer, ForeignKey('master_config.id'))
    reader_id = Column(Integer, ForeignKey('rfidreader_config.id'))
    name = Column(String(MAX_GENERIC_STRING_LENGTH))
    enabled = Column(Boolean)
    measurement_rate_hz = Column(Integer)
    stability_n = Column(Integer)
    tolerance_kg = Column(Float)
    min_mass_kg = Column(Float)
    unlock_mass_kg = Column(Float)
    refload_mass_kg = Column(Float)
    zero_value = Column(Integer)
    refload_value = Column(Integer)
    read_continuously = Column(Boolean)
    amp_signal_filter_mode = Column(Integer)
    fast_response_filter = Column(Boolean)

    # One-to-one relationship:
    reader = relationship("RfidReaderConfig",
                          backref=backref("balance", uselist=False))

    def __init__(self, **kwargs) -> None:
        self.name = kwargs.pop('name', '')
        self.master_config_id = kwargs.pop('master_config_id')
        self.enabled = kwargs.pop('enabled', True)
        self.measurement_rate_hz = kwargs.pop(
            'measurement_rate_hz', DEFAULT_BALANCE_READ_FREQUENCY_HZ)
        self.stability_n = kwargs.pop('stability_n', 5)
        self.tolerance_kg = kwargs.pop('tolerance_kg', 0.005)
        self.min_mass_kg = kwargs.pop('min_mass_kg', 0.050)
        self.unlock_mass_kg = kwargs.pop('unlock_mass_kg', 0.010)
        self.refload_mass_kg = kwargs.pop('refload_mass_kg', 0.1)
        self.read_continuously = kwargs.pop('read_continuously', False)
        self.amp_signal_filter_mode = kwargs.pop('amp_signal_filter_mode', 0)
        self.fast_response_filter = kwargs.pop('fast_response_filter', False)
        kwargs.setdefault('baudrate', 9600)
        kwargs.setdefault('bytesize', serial.EIGHTBITS)
        kwargs.setdefault('parity', serial.PARITY_EVEN)
        kwargs.setdefault('stopbits', serial.STOPBITS_ONE)
        kwargs.setdefault('xonxoff', True)  # p15
        kwargs.setdefault('rtscts', False)
        kwargs.setdefault('dsrdtr', False)
        super().__init__(**kwargs)

    def __repr__(self) -> str:
        return ordered_repr(
            self,
            ['id', 'master_config_id', 'reader_id', 'name',
             'enabled', 'measurement_rate_hz', 'stability_n',
             'tolerance_kg', 'min_mass_kg', 'unlock_mass_kg',
             'refload_mass_kg', 'zero_value', 'refload_value',
             'read_continuously'
             ] + self.fields_component_serial_port())

    def __str__(self) -> str:
        return "{}{}: {}".format(
            "[DISABLED] " if not self.enabled else "",
            self.name or "[no name]",
            self.str_component_serial_port())

    # noinspection PyShadowingBuiltins
    @classmethod
    def get_name_from_id(cls, session: Session, id: int) -> Optional[str]:
        obj = session.query(cls).get(id)
        if obj is None:
            return None
        return obj.name


# =============================================================================
# Program configuration
# =============================================================================

class MasterConfig(SqlAlchemyAttrDictMixin, Base):
    __tablename__ = 'master_config'

    id = Column(Integer, primary_key=True)
    name = Column(String(MAX_GENERIC_STRING_LENGTH))
    modified_at = Column(ArrowMicrosecondType, default=arrow.now,
                         onupdate=arrow.now)
    server = Column(String(MAX_GENERIC_STRING_LENGTH))
    port = Column(Integer)
    wcm_prefix = Column(String(MAX_GENERIC_STRING_LENGTH))
    rfid_effective_time_s = Column(Float)
    rfidreader_configs = relationship("RfidReaderConfig")
    balance_configs = relationship("BalanceConfig")

    def __init__(self, **kwargs) -> None:
        self.server = kwargs.pop('server', 'localhost')
        self.port = kwargs.pop('port', 3233)
        self.wcm_prefix = kwargs.pop('wcm_prefix', 'starfeeder')
        self.rfid_effective_time_s = kwargs.pop('rfid_effective_time_s', 5.0)

    @classmethod
    def get_latest_or_create(cls, session: Session) -> 'MasterConfig':
        config = session.query(cls).order_by(cls.modified_at.desc()).first()
        if config is None:
            config = cls()
            session.add(config)
        return config

    @classmethod
    def get_singleton(cls, session: Session,
                      singleton_pk: int = 1) -> 'MasterConfig':
        config = session.query(cls).get(singleton_pk)
        if config is None:
            config = cls(id=singleton_pk)
            session.add(config)
        return config


# =============================================================================
# RFID detected
# =============================================================================

class RfidEvent(object):
    """
    This is NOT a database object; it's a simple Python object to get passed
    around between threads (we can't use raw Python numbers for this because
    we have to pass around a 64-bit integer).
    It represents a SINGLE detection event of an RFID.
    This is not of much behavioural interest, or interest for recording, since
    dozens of these get generated in a very short space of time.
    What is of more behavioural interest is the RfidEventRecord, below.
    """
    def __init__(self,
                 reader_id: int, reader_name: str,
                 rfid: int, timestamp: arrow.Arrow) -> None:
        self.reader_id = reader_id
        self.reader_name = reader_name
        self.rfid = rfid
        self.timestamp = timestamp

    def __repr__(self) -> str:
        return simple_repr(self)


# =============================================================================
# RFID recording class, glossing over irrelevant duplication
# =============================================================================

class RfidEventRecord(SqlAlchemyAttrDictMixin, Base):
    """
    See rfid.py for a discussion of the RFID tag format.
    Upshot: it'll fit into a 64-bit integer, so we use BigInteger.
    """
    __tablename__ = 'rfid_event'
    id = Column(Integer, primary_key=True)
    reader_id = Column(Integer, ForeignKey('rfidreader_config.id'))
    rfid = Column(BigInteger)
    first_detected_at = Column(ArrowMicrosecondType, default=arrow.now)
    last_detected_at = Column(ArrowMicrosecondType, default=arrow.now)
    n_events = Column(Integer, default=1)

    @classmethod
    def get_ongoing_event_by_reader_rfid(
            cls, session: Session, reader_id: int, rfid: int, now: arrow.Arrow,
            rfid_effective_time_s: float) -> 'RfidEventRecord':
        """Used for RFID recording. Matches reader and RFID."""
        most_recent_of_interest = (
            now - datetime.timedelta(seconds=rfid_effective_time_s)
        )
        return (
            session.query(cls).filter(
                cls.reader_id == reader_id,
                cls.rfid == rfid,
                cls.last_detected_at >= most_recent_of_interest
            ).first()  # There should be no more than one!
        )

    @classmethod
    def record_rfid_detection(
            cls, session: Session, rfid_single_event: RfidEvent,
            rfid_effective_time_s: float) -> None:
        if not isinstance(rfid_single_event, RfidEvent):
            log.critical("Bad rfid_event: {}".format(rfid_single_event))
            return
        reader_id = rfid_single_event.reader_id
        event = cls.get_ongoing_event_by_reader_rfid(
            session,
            reader_id,
            rfid_single_event.rfid,
            rfid_single_event.timestamp,
            rfid_effective_time_s)
        if event:  # one exists already
            event.last_detected_at = rfid_single_event.timestamp
            event.n_events += 1
        else:  # make a new one
            event = cls(
                reader_id=reader_id,
                rfid=rfid_single_event.rfid,
                first_detected_at=rfid_single_event.timestamp,
                last_detected_at=rfid_single_event.timestamp,
                n_events=1,
            )
            session.add(event)
        session.commit()  # ASAP to unlock database


# While a free-floating ID at a particular reader may be of interest,
# a free-floating mass with no ID attached is not of interest.


# =============================================================================
# Raw mass from balance; not a database object
# =============================================================================

class MassEvent(object):
    def __init__(self,
                 balance_id: int, balance_name: str,
                 reader_id: int, reader_name: str,
                 rfid: int, mass_kg: float,
                 timestamp: arrow.Arrow, stable: bool, locked: bool) -> None:
        self.balance_id = balance_id
        self.balance_name = balance_name
        self.reader_id = reader_id
        self.reader_name = reader_name
        self.rfid = rfid
        self.mass_kg = mass_kg
        self.timestamp = timestamp
        self.stable = stable
        self.locked = locked

    def __repr__(self) -> str:
        return simple_repr(self)

    def __str__(self) -> str:
        return "Balance {}: {} kg (RFID: {})".format(
            self.balance_id, self.mass_kg, self.rfid)


# =============================================================================
# Mass detected from a specific RFID-identified individual
# =============================================================================

class MassEventRecord(SqlAlchemyAttrDictMixin, Base):
    __tablename__ = 'mass_event'

    id = Column(Integer, primary_key=True)
    # Who?
    rfid = Column(BigInteger)
    # Where?
    reader_id = Column(Integer, ForeignKey('rfidreader_config.id'))
    balance_id = Column(Integer, ForeignKey('balance_config.id'))
    # When? (Default precision is microseconds, even on SQLite.)
    at = Column(ArrowMicrosecondType, default=arrow.now)
    # ... avoid 'timestamp' as a column name; it's an SQL keyword (adds hassle)
    # How heavy?
    mass_kg = Column(Float)

    @classmethod
    def record_mass_detection(cls, session: Session,
                              mass_single_event: MassEvent) -> None:
        """Returns an OrderedNamespace object, not an SQLAlchemy ORM object."""
        if not isinstance(mass_single_event, MassEvent):
            log.critical("Bad mass_single_event: {}".format(mass_single_event))
            return
        balance_id = mass_single_event.balance_id
        # noinspection PyArgumentList
        mass_identified_event = MassEventRecord(
            rfid=mass_single_event.rfid,
            reader_id=mass_single_event.reader_id,
            balance_id=balance_id,
            at=mass_single_event.timestamp,
            mass_kg=mass_single_event.mass_kg,
        )
        session.add(mass_identified_event)
        session.commit()  # ASAP to unlock database
        # return mass_identified_event.get_attrdict()



#!/usr/bin/env python
# starfeeder/constants.py

"""
    Copyright (C) 2015-2015 Rudolf Cardinal (rudolf@pobox.com).

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from enum import Enum
import os
import sys
from starfeeder.version import VERSION


ABOUT = """
<b>Starfeeder {VERSION}</b><br>
<br>
Whisker bird monitor.<br>
<br>
Functions:
<ul>
  <li>
    Talks to
    <ul>
      <li>multiple radiofrequency identification (RFID) readers</li>
      <li>multiple weighing balances</li>
      <li>one Whisker server (<a href="{WHISKER_URL}">{WHISKER_URL}</a>)</li>
    </ul>
  </li>
  <li>Detects the mass of subjects identified by their RFID (having configured
    RFID readers/balances into pairs)</li>
  <li>Tells the Whisker server, and its other clients, about RFID and mass
    events.</li>
  <li>Stores its data to a database (e.g. SQLite; MySQL).</li>
</ul>

Hardware supported:
<ul>
  <li>RFID readers: MBRose BW-1/001 integrated load cell scale RFID reader</li>
  <li>Balances: HBM AD105 digital transducer with RS-485 interface</li>
  <li>RFID readers and balances connect to a USB controller (MBRose WS-2/001
    RS232-to-USB interface), which connects to the computer's USB port.</li>
</ul>

You will also need:
<ul>
  <li>A database. Any backend supported by SQLAlchemy will do (see
    <a href="{BACKEND_URL}">{BACKEND_URL}</a>).
    SQLite is quick. Starfeeder finds its database using the environment
    variable STARFEEDER_DATABASE_URL.</li>
  <li>You may want a graphical tool for database management. There are lots.
    For SQLite, consider Sqliteman
    (<a href="{SQLITEMAN_URL}">{SQLITEMAN_URL}</a>).
</ul>

By Rudolf Cardinal (rudolf@pobox.com).<br>
Copyright &copy; 2015-2016 Rudolf Cardinal.
For licensing details see LICENSE.txt.<br>
External libraries used include Alembic; bitstring; PyInstaller; PySerial;
Qt (via PyQt5); SQLAlchemy.
""".format(
    VERSION=VERSION,
    WHISKER_URL="http://www.whiskercontrol.com/",
    SQLITEMAN_URL="http://sqliteman.yarpen.cz/",
    BACKEND_URL="http://docs.sqlalchemy.org/en/latest/core/engines.html",
)

BALANCE_ASF_MINIMUM = 0  # p37 of balance manual
BALANCE_ASF_MAXIMUM = 8  # p37 of balance manual

DEFAULT_BALANCE_READ_FREQUENCY_HZ = 6

DB_URL_ENV_VAR = "STARFEEDER_DATABASE_URL"
DATABASE_ENV_VAR_NOT_SPECIFIED = """
===============================================================================
You must specify the {var} environment variable (which is an
SQLAlchemy database URL). Examples follow.

Windows:
    set {var}=sqlite:///C:\\path\\to\\database.sqlite3
Linux:
    export {var}=sqlite:////absolute/path/to/database.sqlite3
===============================================================================
""".format(var=DB_URL_ENV_VAR)
GUI_MASS_FORMAT = '% 9.6f'
GUI_TIME_FORMAT = '%H:%M:%S'
LOG_FORMAT = '%(asctime)s.%(msecs)03d:%(levelname)s:%(name)s:%(message)s'
LOG_DATEFMT = '%Y-%m-%d %H:%M:%S'
WINDOW_TITLE = 'Starfeeder: RFID/balance controller for Whisker'
WRONG_DATABASE_VERSION_STUB = """
===============================================================================
Database revision should be {head_revision} but is {current_revision}.

- If the database version is too low, run starfeeder with the
  "--upgrade-database" parameter (because your database is too old), or click
  the "Upgrade database" button in the GUI.

- If the database version is too high, upgrade starfeeder (because you're
  trying to use an old starfeeder version with a newer database).
===============================================================================
"""

# =============================================================================
# Find out where Alembic and other files live
# =============================================================================

if getattr(sys, 'frozen', False):
    # Running inside a PyInstaller bundle.
    # http://pythonhosted.org/PyInstaller/#run-time-operation
    # noinspection PyProtectedMember,PyUnresolvedReferences
    CURRENT_DIR = sys._MEIPASS
else:
    # Running in a normal Python environment.
    CURRENT_DIR = os.path.dirname(os.path.abspath(__file__))

ALEMBIC_CONFIG_FILENAME = os.path.join(CURRENT_DIR, 'alembic.ini')
ALEMBIC_BASE_DIR = CURRENT_DIR
MANUAL_FILENAME = os.path.join(CURRENT_DIR, 'manual.pdf')


# =============================================================================
# Thread state enum
# =============================================================================

class ThreadOwnerState(Enum):
    stopped = 0
    starting = 1
    running = 2
    stopping = 3



#!/usr/bin/env/python
# starfeeder/common_serial_testing.py

from argparse import ArgumentParser, Namespace
import logging
import random
import re
from typing import Optional, Pattern, Sequence, Union

import serial

from starfeeder.serial_controller import LF, READ_TIMEOUT_SEC

log = logging.getLogger(__name__)


# =============================================================================
# Generic command-line arguments for PySerial serial ports
# =============================================================================

def add_serial_port_args(parser: ArgumentParser) -> None:
    group = parser.add_argument_group('serial port')
    group.add_argument(
        '--port', type=str, required=True,
        help="Serial port (e.g. Linux: /dev/ttyS0, Windows: COM1). For Linux "
             "testing, you can use 'socat -d -d [-d [-d]] "
             "pty,raw,echo=0,b9600 pty,raw,echo=0,b9600' to create a pair of "
             "linked pseudo-ports, and use the resulting /dev/pts/* devices,"
             "one for this program, and the other for the program you wish "
             "to test. For manual testing, you can replace one of the socat "
             "settings with '-' and type into it, but that only works with "
             "LF (and similar) line endings. For raw manual testing (e.g. for "
             "devices that expect commands without CR/LF endings), use socat "
             "as above and talk to the other port with e.g. 'minicom "
             "--device /dev/pts/SOMETHING'.")
    group.add_argument(
        '--baudrate', nargs='?', type=int, default=9600,
        choices=serial.SerialBase.BAUDRATES,
        help="Baud rate")
    group.add_argument(
        '--bytesize', nargs='?', type=int, default=serial.EIGHTBITS,
        choices=serial.SerialBase.BYTESIZES,
        help="Byte size")
    group.add_argument(
        '--parity', nargs='?',
        type=lambda c: c.upper(),
        default=serial.PARITY_NONE,
        choices=serial.SerialBase.PARITIES,
        help="Parity")
    group.add_argument(
        '--stopbits', nargs='?', type=float, default=serial.STOPBITS_ONE,
        choices=serial.SerialBase.STOPBITS,
        help="Stop bits")
    # group.add_argument(
    #     '--timeout', type=float, default=None,
    #     help="Read timeout (s)")
    group.add_argument(
        '--xonxoff', action='store_true',
        help="Use XON/XOFF software flow control")
    group.add_argument(
        '--rtscts', action='store_true',
        help="Use RTS/CTS hardware flow control")
    group.add_argument(
        '--rts', type=int, default=None,
        choices=[0, 1],
        help='Set initial RTS line state')
    group.add_argument(
        '--dsrdtr', action='store_true',
        help="Use DSR/DTR hardware flow control")
    group.add_argument(
        '--dtr', type=int, default=None,
        choices=[0, 1],
        help='Set initial DTR line state')
    group.add_argument(
        '--write_timeout', type=float, default=None,
        help="Write timeout (s)")
    group.add_argument(
        '--inter_byte_timeout', type=float, default=None,
        help="Inter-byte timeout (s)")


# =============================================================================
# Split by multiple delimiters
# =============================================================================
# http://stackoverflow.com/questions/1059559/python-split-strings-with-multiple-delimiters  # noqa

REGEX_METACHARS = ["\\", "^", "$", ".",
                   "|", "?", "*", "+",
                   "(", ")", "[", "{"]
# http://www.regular-expressions.info/characters.html
# Start with \, for replacement.


def escape_literal_string_for_regex(s: str) -> str:
    r"""
    Escape any regex characters.

    Start with \ -> \\
        ... this should be the first replacement in REGEX_METACHARS.
    """
    for c in REGEX_METACHARS:
        s.replace(c, "\\" + c)
    return s


def escape_literal_bytes_for_regex(b: bytes) -> bytes:
    for c in REGEX_METACHARS:
        cb = c.encode("ascii")
        b.replace(cb, b"\\" + cb)
    return b


def compile_literal_splitter_regex(
        separators: Sequence[Union[str, bytes]],
        flags: int = re.IGNORECASE) -> Pattern:
    if not separators:
        raise ValueError("No separators provided")
    if isinstance(separators[0], bytes):
        escaped_separators = [b'(?:' + escape_literal_bytes_for_regex(s) + b')'
                              for s in separators]
        regex_str = b'|'.join(escaped_separators)
    else:
        escaped_separators = ['(?:' + escape_literal_string_for_regex(s) + ')'
                              for s in separators]
        regex_str = '|'.join(escaped_separators)
    log.debug("separators: {}".format(repr(separators)))
    log.debug("regex_str: {}".format(repr(regex_str)))
    return re.compile(regex_str, flags)


# =============================================================================
# Class to process serial port data from the command line
# =============================================================================

class CommandLineSerialProcessor(object):
    # noinspection PyUnresolvedReferences
    def __init__(self,
                 args: Namespace,
                 inbound_eol: Optional[Union[bytes, Sequence[bytes]]] = LF,
                 inbound_bytewise: bool = False,
                 outbound_eol: bytes = LF,
                 encoding: str = 'ascii',
                 read_timeout_sec: float = READ_TIMEOUT_SEC,
                 **kwargs) -> None:
        super().__init__(**kwargs)
        self.serial_port = serial.Serial(
            port=args.port,
            baudrate=args.baudrate,
            bytesize=args.bytesize,
            parity=args.parity,
            stopbits=args.stopbits,
            timeout=read_timeout_sec,
            xonxoff=args.xonxoff,
            rtscts=args.rtscts,
            dsrdtr=args.dsrdtr,
            write_timeout=args.write_timeout,
            inter_byte_timeout=args.inter_byte_timeout
        )
        if args.rts is not None:
            self.serial_port.rts = args.rts
        if args.dtr is not None:
            self.serial_port.dtr = args.dtr

        if isinstance(inbound_eol, str):
            self.inbound_eol = [inbound_eol]
        elif not inbound_eol:
            self.inbound_eol = []
        else:
            self.inbound_eol = inbound_eol
            if len(self.inbound_eol) > 1:
                # noinspection PyTypeChecker
                for individual_eol in self.inbound_eol:
                    if not individual_eol or len(individual_eol) == 0:
                        raise ValueError(
                            "Can't have any zero-length inbound_eol members if"
                            " some have non-zero length")
        if self.inbound_eol:
            self.inbound_eol_splitter = compile_literal_splitter_regex(
                self.inbound_eol)
        else:
            self.inbound_eol_splitter = None
        self.use_inbound_eol = inbound_eol and len(inbound_eol[0]) > 0
        self.inbound_bytewise = inbound_bytewise
        self.outbound_eol = outbound_eol
        self.encoding = encoding
        self.residual = b''

    def run(self) -> None:
        self.start()
        log.info("Starting to read from serial port")
        while True:
            data = self.serial_port.read(1)  # will wait until timeout
            # ... will return b'' if no data
            if not self.inbound_bytewise:
                data += self.serial_port.read(self.serial_port.inWaiting())
            if len(data) > 0:
                self.bytes_received(data)
            self.spontaneous()

    def start(self) -> None:
        """May be overridden."""
        pass

    def spontaneous(self) -> None:
        """May be overridden."""
        pass

    def bytes_received(self, data: bytes) -> None:
        """May be overridden."""
        log.debug("bytes_received: {}".format(repr(data)))
        if self.inbound_bytewise or not self.inbound_eol_splitter:
            try:
                decoded = data.decode(self.encoding)
            except UnicodeDecodeError:
                log.critical("Can't decode using {} encoding; data was "
                             "{}".format(repr(data)))
                return
            self.line_received(decoded)
        else:
            data = self.residual + data
            fragments = self.inbound_eol_splitter.split(data)
            lines = fragments[:-1]
            self.residual = fragments[-1]
            # log.debug("lines={}, residual={}".format(
            #     repr(lines), repr(self.residual)))
            for line in lines:
                try:
                    decoded = line.decode(self.encoding)
                except UnicodeDecodeError:
                    log.critical("Can't decode using {} encoding; data was "
                                 "{}".format(repr(data)))
                    return
                self.line_received(decoded)

    # noinspection PyMethodMayBeStatic
    def line_received(self, line: str) -> None:
        log.info("line_received: {}".format(repr(line)))

    def send_line(self, line: str) -> None:
        log.info("sending line: {}".format(repr(line)))
        data = line.encode(self.encoding) + self.outbound_eol
        self.send_bytes(data)

    def send_bytes(self, data: bytes) -> None:
        log.debug("sending data: {}".format(repr(data)))
        self.serial_port.write(data)


# =============================================================================
# Flip a coin
# =============================================================================

def coin(p: float) -> bool:
    return random.random() < p



#!/usr/bin/env python
# starfeeder/gui.py

"""
    Copyright (C) 2015-2015 Rudolf Cardinal (rudolf@pobox.com).

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

import collections
import logging
import platform
import traceback
from typing import List

import arrow
from PyQt5.QtCore import QObject, Qt, pyqtSignal, pyqtSlot
from PyQt5.QtGui import (
    QCloseEvent,  # for type hints
)
from PyQt5.QtWidgets import (
    QApplication,
    QCheckBox,
    QComboBox,
    QDialog,
    QDialogButtonBox,
    QFormLayout,
    QGridLayout,
    QHBoxLayout,
    QLabel,
    QLineEdit,
    QMainWindow,
    QMessageBox,
    QPushButton,
    QVBoxLayout,
    QWidget,
)
import serial
from serial.tools.list_ports import comports
from sqlalchemy.sql import exists
from sqlalchemy.orm.session import Session  # for type hints
from whisker.lang import launch_external_file, natural_keys
from whisker.qt import (
    exit_on_exception,
    GenericListModel,
    ModalEditListView,
    RadioGroup,
    StyledQGroupBox,
    TextLogElement,
    TransactionalEditDialogMixin,
    ValidationError,
)
from whisker.qtclient import WhiskerOwner
from whisker.sqlalchemy import (
    database_is_sqlite,
    session_thread_scope,
    upgrade_database,
)

from starfeeder.balance import BalanceOwner, RATE_MAP_HZ_TO_CODE
from starfeeder.constants import (
    ABOUT,
    ALEMBIC_BASE_DIR,
    ALEMBIC_CONFIG_FILENAME,
    BALANCE_ASF_MINIMUM,
    BALANCE_ASF_MAXIMUM,
    DATABASE_ENV_VAR_NOT_SPECIFIED,
    GUI_MASS_FORMAT,
    GUI_TIME_FORMAT,
    # HELP,
    MANUAL_FILENAME,
    WINDOW_TITLE,
    WRONG_DATABASE_VERSION_STUB,
)
from starfeeder.models import (
    BalanceConfig,
    CalibrationReport,  # for type hints
    MassEvent,  # for type hints
    MassEventRecord,
    MasterConfig,
    RfidEvent,  # for type hints
    RfidEventRecord,
    RfidReaderConfig,
    SerialPortConfigMixin,  # for type hints
)
from starfeeder.rfid import RfidOwner
from starfeeder.settings import get_database_settings
from starfeeder.task import WeightWhiskerTask

log = logging.getLogger(__name__)


# =============================================================================
# Constants
# =============================================================================

# POSSIBLE_RATES_HZ = [100, 50, 25, 10, 6, 3, 2, 1]
POSSIBLE_RATES_HZ = [12, 6, 3, 2, 1]
# ... 100 Hz (a) ends up with a bunch of messages concatenated from the serial
# device, so timing becomes pointless, (b) is pointless, and (c) leads rapidly
# to a segmentation fault.
# Note that 9600 bps at 8E1 = 960 cps.
# So divide that by the length of the message (including CR+LF) to get the
# absolute maximum rate. And don't go near that.
if any([x not in RATE_MAP_HZ_TO_CODE.keys() for x in POSSIBLE_RATES_HZ]):
    raise AssertionError("Invalid frequency in POSSIBLE_RATES_HZ")

POSSIBLE_ASF_MODES = list(range(BALANCE_ASF_MINIMUM, BALANCE_ASF_MAXIMUM + 1))

ALIGNMENT = Qt.AlignLeft | Qt.AlignTop
DEVICE_ID_LABEL = "Device ID (set when first saved)"
RENAME_WARNING = (
    "<b>Once created and used for real data, AVOID RENAMING devices;<br>"
    "RFID/mass data will refer to these entries by number (not name).</b>"
)


# =============================================================================
# Secondary GUI windows
# =============================================================================

class NoDatabaseSpecifiedWindow(QDialog):
    exit_kill_log = pyqtSignal()

    # noinspection PyUnresolvedReferences
    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)
        self.setWindowTitle(WINDOW_TITLE)
        info = QLabel(DATABASE_ENV_VAR_NOT_SPECIFIED)
        ok_buttons = QDialogButtonBox(QDialogButtonBox.Ok,
                                      Qt.Horizontal, self)
        ok_buttons.accepted.connect(self.exit_kill_log)
        ok_buttons.accepted.connect(self.accept)
        layout = QVBoxLayout()
        layout.addWidget(info)
        layout.addWidget(ok_buttons)
        self.setLayout(layout)

    # noinspection PyPep8Naming
    def closeEvent(self, event: QCloseEvent) -> None:
        self.exit_kill_log.emit()
        event.accept()


class WrongDatabaseVersionWindow(QDialog):
    exit_kill_log = pyqtSignal()

    # noinspection PyUnresolvedReferences
    def __init__(self, current_revision: str, head_revision: str,
                 **kwargs) -> None:
        super().__init__(**kwargs)
        self.setWindowTitle(WINDOW_TITLE)

        info = QLabel(WRONG_DATABASE_VERSION_STUB.format(
            head_revision=head_revision,
            current_revision=current_revision))
        upgrade_button = QPushButton("Upgrade database")
        upgrade_button.clicked.connect(self.upgrade_database)
        ok_buttons = QDialogButtonBox(QDialogButtonBox.Ok,
                                      Qt.Horizontal, self)
        ok_buttons.accepted.connect(self.exit_kill_log)
        ok_buttons.accepted.connect(self.accept)

        layout_upgrade = QHBoxLayout()
        layout_upgrade.addWidget(upgrade_button)
        layout_upgrade.addStretch(1)
        main_layout = QVBoxLayout()
        main_layout.addWidget(info)
        main_layout.addLayout(layout_upgrade)
        main_layout.addWidget(ok_buttons)
        self.setLayout(main_layout)

    @pyqtSlot()
    def upgrade_database(self) -> None:
        try:
            upgrade_database(ALEMBIC_CONFIG_FILENAME, ALEMBIC_BASE_DIR)
            # noinspection PyCallByClass
            QMessageBox.about(self, "Success",
                              "Successfully upgraded database.")
        except Exception as e:  # catch any database upgrade failure
            # noinspection PyCallByClass
            QMessageBox.about(
                self, "Failure",
                "Failed to upgrade database. Error was: {}.\n\n{}".format(
                    str(e), traceback.format_exc()))

    # noinspection PyPep8Naming
    def closeEvent(self, event: QCloseEvent) -> None:
        self.exit_kill_log.emit()
        event.accept()


# =============================================================================
# Main GUI window
# =============================================================================

class BaseWindow(QMainWindow):  # GUI thread
    # Don't inherit from QDialog, which has an additional Escape-to-close
    # function that's harder to trap. Use QWidget or QMainWindow.
    NAME = "main"

    exit_kill_log = pyqtSignal()

    # noinspection PyUnresolvedReferences
    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)
        self.exit_pending = False

        # ---------------------------------------------------------------------
        # Internals
        # ---------------------------------------------------------------------
        self.rfid_list = []
        self.balance_list = []
        self.whisker_task = None
        self.whisker_owner = None
        self.dbsettings = get_database_settings()

        # ---------------------------------------------------------------------
        # GUI
        # ---------------------------------------------------------------------
        self.setWindowTitle(WINDOW_TITLE)
        self.setMinimumWidth(400)

        config_group = StyledQGroupBox("Configure")
        config_layout = QHBoxLayout()
        self.configure_button = QPushButton('&Configure')
        self.configure_button.clicked.connect(self.configure)
        self.calibrate_balances_button = QPushButton(
            '&Tare/calibrate balances')
        self.calibrate_balances_button.clicked.connect(self.calibrate_balances)
        config_layout.addWidget(self.configure_button)
        config_layout.addWidget(self.calibrate_balances_button)
        config_layout.addStretch(1)
        config_group.setLayout(config_layout)

        run_group = StyledQGroupBox("Run")
        run_layout = QHBoxLayout()
        self.start_button = QPushButton('St&art/reset everything')
        self.start_button.clicked.connect(self.start)
        self.stop_button = QPushButton('Sto&p everything')
        self.stop_button.clicked.connect(self.stop)
        run_layout.addWidget(self.start_button)
        run_layout.addWidget(self.stop_button)
        run_layout.addStretch(1)
        run_group.setLayout(run_layout)

        test_group = StyledQGroupBox("Testing and information")
        test_layout = QHBoxLayout()
        self.reset_rfids_button = QPushButton('Reset RFIDs')
        self.reset_rfids_button.clicked.connect(self.reset_rfid_devices)
        self.ping_balances_button = QPushButton('Ping &balances')
        self.ping_balances_button.clicked.connect(self.ping_balances)
        self.ping_whisker_button = QPushButton('&Ping Whisker')
        self.ping_whisker_button.clicked.connect(self.ping_whisker)
        report_status_button = QPushButton('&Report status')
        report_status_button.clicked.connect(self.report_status)
        help_button = QPushButton('&Help')
        help_button.clicked.connect(self.help)
        about_button = QPushButton('&About')
        about_button.clicked.connect(self.about)
        test_layout.addWidget(self.reset_rfids_button)
        test_layout.addWidget(self.ping_balances_button)
        test_layout.addWidget(self.ping_whisker_button)
        test_layout.addWidget(report_status_button)
        test_layout.addWidget(help_button)
        test_layout.addWidget(about_button)
        test_layout.addStretch(1)
        test_group.setLayout(test_layout)

        self.status_group = StyledQGroupBox("Status")
        self.status_layout = None
        self.whisker_label_server = None
        self.whisker_label_port = None
        self.whisker_label_status = None
        self.lay_out_status("-", "-", "-")

        self.log = TextLogElement()

        # You can't use the layout as the parent of the widget.
        # But you don't need to specify a parent when you use addWidget; it
        # works that out for you.
        # http://doc.qt.io/qt-4.8/layout.html#tips-for-using-layouts
        main_widget = QWidget(self)
        self.setCentralWidget(main_widget)
        main_layout = QVBoxLayout(main_widget)
        main_layout.addWidget(config_group)
        main_layout.addWidget(run_group)
        main_layout.addWidget(test_group)
        main_layout.addWidget(self.status_group)
        main_layout.addWidget(self.log.get_widget())

        self.set_button_states()

        self.rfid_list = []
        self.rfid_id_to_obj = {}
        self.rfid_id_to_idx = {}
        self.balance_list = []
        self.balance_id_to_obj = {}
        self.balance_id_to_idx = {}
        self.balance_labels_status = []
        self.balance_labels_raw_mass = []
        self.balance_labels_raw_mass_at = []
        self.balance_labels_stable_mass = []
        self.balance_labels_stable_mass_at = []
        self.balance_labels_idmass = []
        self.balance_labels_rfid = []
        self.balance_labels_idmass_at = []
        self.rfid_labels_status = []
        self.rfid_labels_rfid = []
        self.rfid_labels_at = []

    # -------------------------------------------------------------------------
    # Exiting
    # -------------------------------------------------------------------------

    # noinspection PyPep8Naming
    def closeEvent(self, event: QCloseEvent) -> None:
        """Trap exit."""
        quit_msg = "Are you sure you want to exit?"
        # noinspection PyCallByClass
        reply = QMessageBox.question(self, 'Really exit?', quit_msg,
                                     QMessageBox.Yes, QMessageBox.No)
        if reply != QMessageBox.Yes:
            event.ignore()
            return
        # If subthreads aren't shut down, we get a segfault when we quit.
        # However, right now, signals aren't being processed because we're in
        # the GUI message loop. So we need to defer the call if subthreads are
        # running
        if not self.anything_running():
            self.exit_kill_log.emit()
            event.accept()  # actually quit
            return
        # Now stop everything
        log.warning("Waiting for threads to finish...")
        self.exit_pending = True
        for rfid in self.rfid_list:
            rfid.stop()
        for balance in self.balance_list:
            balance.stop()
        if self.whisker_owner:
            self.whisker_owner.stop()
        # Will get one or more callbacks to something_finished
        event.ignore()

    # -------------------------------------------------------------------------
    # Configuration
    # -------------------------------------------------------------------------

    @pyqtSlot()
    def configure(self) -> None:
        readonly = self.anything_running()
        with session_thread_scope(self.dbsettings, readonly) as session:
            config = MasterConfig.get_singleton(session)
            dialog = MasterConfigWindow(session, config, parent=self,
                                        readonly=readonly)
            dialog.edit_in_nested_transaction()
            session.commit()

    # -------------------------------------------------------------------------
    # Starting, stopping, thread management
    # -------------------------------------------------------------------------

    @pyqtSlot()
    def start(self) -> None:
        if self.anything_running():
            # noinspection PyCallByClass
            QMessageBox.about(self, "Can't start",
                              "Can't start: already running.")
            return

        with session_thread_scope(self.dbsettings) as session:
            config = MasterConfig.get_singleton(session)
            # Continue to hold the session beyond this.
            # http://stackoverflow.com/questions/13904735/sqlalchemy-how-to-use-an-instance-in-sqlalchemy-after-session-close  # noqa

            # -----------------------------------------------------------------
            # Whisker
            # -----------------------------------------------------------------
            self.whisker_task = WeightWhiskerTask(wcm_prefix=config.wcm_prefix)
            self.whisker_owner = WhiskerOwner(  # GUI thread
                self.whisker_task, config.server, parent=self)
            self.whisker_owner.finished.connect(self.something_finished)
            self.whisker_owner.status_sent.connect(self.on_status)
            self.whisker_owner.error_sent.connect(self.on_status)
            self.whisker_owner.connected.connect(
                self.on_whisker_state_connected)
            self.whisker_owner.disconnected.connect(
                self.on_whisker_state_disconnected)
            self.whisker_owner.finished.connect(
                self.on_whisker_state_disconnected)
            self.whisker_owner.pingack_received.connect(
                self.on_whisker_pingack_received)
            # It's OK to connect signals before or after moving them to a
            # different thread: http://stackoverflow.com/questions/20752154
            # We don't want time-critical signals going via the GUI thread,
            # because that might be busy with user input.
            # So we'll use the self.whisker_task as the recipient; see below.

            # -----------------------------------------------------------------
            # RFIDs
            # -----------------------------------------------------------------
            self.rfid_list = []
            self.rfid_id_to_obj = {}
            self.rfid_id_to_idx = {}
            for i, rfid_config in enumerate(config.rfidreader_configs):
                if not rfid_config.enabled:
                    continue
                rfid = RfidOwner(rfid_config, callback_id=i, parent=self)
                rfid.status_sent.connect(self.on_status)
                rfid.error_sent.connect(self.on_status)
                rfid.finished.connect(self.something_finished)
                rfid.rfid_received.connect(self.whisker_task.on_rfid)  # different thread  # noqa
                rfid.rfid_received.connect(self.on_rfid)
                rfid.state_change.connect(self.on_rfid_state)
                self.rfid_list.append(rfid)
                self.rfid_id_to_obj[rfid.reader_id] = rfid
                self.rfid_id_to_idx[rfid.reader_id] = i

            # -----------------------------------------------------------------
            # Balances
            # -----------------------------------------------------------------
            self.balance_list = []
            self.balance_id_to_obj = {}
            self.balance_id_to_idx = {}
            for i, balance_config in enumerate(config.balance_configs):
                if not balance_config.enabled:
                    continue
                if not balance_config.reader:
                    continue
                if not balance_config.reader.enabled:
                    continue
                balance = BalanceOwner(
                    balance_config,
                    rfid_effective_time_s=config.rfid_effective_time_s,
                    callback_id=i,
                    parent=self)
                balance.status_sent.connect(self.on_status)
                balance.error_sent.connect(self.on_status)
                balance.finished.connect(self.something_finished)
                balance.mass_received.connect(self.whisker_task.on_mass)  # different thread  # noqa
                balance.mass_received.connect(self.on_mass)  # same thread
                balance.calibrated.connect(self.on_calibrated)
                balance.state_change.connect(self.on_balance_state)
                self.balance_list.append(balance)
                self.balance_id_to_obj[balance.balance_id] = balance
                self.balance_id_to_idx[balance.balance_id] = i
                rfid = self.rfid_id_to_obj[balance_config.reader_id]
                rfid.rfid_received.connect(balance.on_rfid)

            # -----------------------------------------------------------------
            # Display
            # -----------------------------------------------------------------
            self.lay_out_status(config.server, str(config.port),
                                "Not connected")

        # ---------------------------------------------------------------------
        # Start
        # ---------------------------------------------------------------------
        for rfid in self.rfid_list:
            rfid.start()
        for balance in self.balance_list:
            balance.start()
        if self.whisker_owner:
            self.whisker_owner.start()
        self.set_button_states()

    @pyqtSlot()
    def stop(self) -> None:
        if not self.anything_running():
            # noinspection PyCallByClass
            QMessageBox.about(self, "Can't stop",
                              "Nothing to stop: not running.")
            return
        self.status("Stopping everything...")
        for rfid in self.rfid_list:
            rfid.stop()
        for balance in self.balance_list:
            balance.stop()
        if self.whisker_owner:
            self.whisker_owner.stop()
        self.set_button_states()

    @pyqtSlot()
    def something_finished(self) -> None:
        if self.anything_running():
            log.debug("... thread finished, but others are still running")
            return
        self.status("All tasks and threads stopped")
        if self.exit_pending:
            self.exit_kill_log.emit()
            # noinspection PyArgumentList
            QApplication.quit()
        self.set_button_states()

    def anything_running(self) -> bool:
        """Returns a bool."""
        return (
            any(r.is_running() for r in self.rfid_list) or
            any(b.is_running() for b in self.balance_list) or
            (self.whisker_owner is not None and
                self.whisker_owner.is_running())
        )

    # -------------------------------------------------------------------------
    # Testing
    # -------------------------------------------------------------------------

    @pyqtSlot()
    def reset_rfid_devices(self) -> None:
        for rfid in self.rfid_list:
            rfid.reset()

    @pyqtSlot()
    def ping_balances(self) -> None:
        for balance in self.balance_list:
            balance.ping()

    @pyqtSlot()
    def calibrate_balances(self) -> None:
        dialog = CalibrateBalancesWindow(balance_owners=self.balance_list,
                                         parent=self)
        dialog.exec_()

    @pyqtSlot()
    def ping_whisker(self) -> None:
        if self.whisker_owner:
            self.whisker_owner.ping()

    @pyqtSlot()
    def report_status(self) -> None:
        self.status("Requesting status from RFID devices")
        for rfid in self.rfid_list:
            rfid.report_status()
        self.status("Requesting status from balances")
        for balance in self.balance_list:
            balance.report_status()
        if self.whisker_owner:
            self.status("Requesting status from Whisker controller")
            self.whisker_owner.report_status()
            # self.whisker_task.report_status()
        self.status("Status report complete.")

    @pyqtSlot()
    def about(self) -> None:
        # noinspection PyCallByClass
        QMessageBox.about(self, "Starfeeder", ABOUT)

    @pyqtSlot()
    def help(self) -> None:
        launch_external_file(MANUAL_FILENAME)
        self.status("Launched {}".format(MANUAL_FILENAME))
        # QMessageBox.about(self, "Starfeeder", HELP)

    # -------------------------------------------------------------------------
    # Calibration
    # -------------------------------------------------------------------------

    def on_calibrated(self, calibration_report: CalibrationReport) -> None:
        msg = str(calibration_report)
        self.status(msg)
        log.info(msg)
        with session_thread_scope(self.dbsettings) as session:
            balance_config = session.query(BalanceConfig).get(
                calibration_report.balance_id)
            log.debug("WAS: {}".format(repr(balance_config)))
            balance_config.zero_value = calibration_report.zero_value
            balance_config.refload_value = calibration_report.refload_value
            log.debug("NOW: {}".format(repr(balance_config)))
            session.commit()

    # -------------------------------------------------------------------------
    # Status log
    # -------------------------------------------------------------------------

    @pyqtSlot(str, str)
    def on_status(self, msg: str, source: str = "") -> None:
        # http://stackoverflow.com/questions/16568451
        if source:
            msg = "[{}] {}".format(source, msg)
        self.log.add(msg)

    def status(self, msg: str) -> None:
        self.on_status(msg, self.NAME)

    # -------------------------------------------------------------------------
    # More GUI
    # -------------------------------------------------------------------------

    def lay_out_status(self, server_info: str = '', port_info: str = '',
                       status_info: str = '') -> None:
        # Since we want to remove and add items, the simplest thing isn't to
        # own the grid layout and remove/add widgets, but to own the Group
        # within which the layout sits, and assign a new layout (presumably
        # garbage-collecting the old ones).
        # Actually, we have to pass ownership of the old layout to a dummy
        # widget owner that's then destroyed;
        # http://stackoverflow.com/questions/10416582/replacing-layout-on-a-qwidget-with-another-layout  # noqa

        # Wipe the old stuff
        if self.status_layout:
            QWidget().setLayout(self.status_layout)
            # ... will deal with the children
        # Now we should be able to redo it:
        rfid_status_grid = QGridLayout()
        rfid_hlayout = QHBoxLayout()
        rfid_hlayout.addLayout(rfid_status_grid)
        rfid_hlayout.addStretch(1)
        balance_status_grid = QGridLayout()
        balance_hlayout = QHBoxLayout()
        balance_hlayout.addLayout(balance_status_grid)
        balance_hlayout.addStretch(1)
        whisker_status_grid = QGridLayout()
        whisker_hlayout = QHBoxLayout()
        whisker_hlayout.addLayout(whisker_status_grid)
        whisker_hlayout.addStretch(1)
        self.status_layout = QVBoxLayout()
        self.status_layout.addLayout(rfid_hlayout)
        self.status_layout.addLayout(balance_hlayout)
        self.status_layout.addLayout(whisker_hlayout)
        self.status_group.setLayout(self.status_layout)

        rfid_status_grid.addWidget(QLabel("<b>RFID reader</b>"),
                                   0, 0, ALIGNMENT)
        rfid_status_grid.addWidget(QLabel("<b>Status</b>"), 0, 1, ALIGNMENT)
        rfid_status_grid.addWidget(QLabel("<b>Last RFID seen</b>"),
                                   0, 2, ALIGNMENT)
        rfid_status_grid.addWidget(QLabel("<b>At</b>"), 0, 3, ALIGNMENT)
        self.rfid_labels_status = []
        self.rfid_labels_rfid = []
        self.rfid_labels_at = []
        for row, rfid in enumerate(self.rfid_list, start=1):
            rfid_status_grid.addWidget(
                QLabel("{}: {}".format(rfid.reader_id, rfid.name)),
                row, 0, ALIGNMENT)
            label = QLabel("-")
            rfid_status_grid.addWidget(label, row, 1, ALIGNMENT)
            self.rfid_labels_status.append(label)
            rfid_label_rfid = QLabel("-")
            rfid_status_grid.addWidget(rfid_label_rfid, row, 2, ALIGNMENT)
            self.rfid_labels_rfid.append(rfid_label_rfid)
            rfid_label_at = QLabel("-")
            rfid_status_grid.addWidget(rfid_label_at, row, 3, ALIGNMENT)
            self.rfid_labels_at.append(rfid_label_at)

        balance_status_grid.addWidget(QLabel("<b>Balance</b>"),
                                      0, 0, ALIGNMENT)
        balance_status_grid.addWidget(QLabel("<b>Status</b>"), 0, 1, ALIGNMENT)
        balance_status_grid.addWidget(QLabel("<b>Raw (kg)</b>"),
                                      0, 2, ALIGNMENT)
        balance_status_grid.addWidget(QLabel("<b>At</b>"),
                                      0, 3, ALIGNMENT)
        balance_status_grid.addWidget(QLabel("<b>Stable (kg)</b>"),
                                      0, 4, ALIGNMENT)
        balance_status_grid.addWidget(QLabel("<b>At</b>"), 0, 5, ALIGNMENT)
        balance_status_grid.addWidget(QLabel("<b>Locked/ID'd (kg)</b>"),
                                      0, 6, ALIGNMENT)
        balance_status_grid.addWidget(QLabel("<b>RFID</b>"), 0, 7, ALIGNMENT)
        balance_status_grid.addWidget(QLabel("<b>At</b>"), 0, 8, ALIGNMENT)
        self.balance_labels_status = []
        self.balance_labels_raw_mass = []
        self.balance_labels_raw_mass_at = []
        self.balance_labels_stable_mass = []
        self.balance_labels_stable_mass_at = []
        self.balance_labels_idmass = []
        self.balance_labels_rfid = []
        self.balance_labels_idmass_at = []
        for row, balance in enumerate(self.balance_list, start=1):
            balance_status_grid.addWidget(
                QLabel("{}: {}".format(balance.balance_id, balance.name)),
                row, 0, ALIGNMENT)
            label = QLabel("-")
            balance_status_grid.addWidget(label, row, 1, ALIGNMENT)
            self.balance_labels_status.append(label)
            label = QLabel("-")
            balance_status_grid.addWidget(label, row, 2, ALIGNMENT)
            self.balance_labels_raw_mass.append(label)
            label = QLabel("-")
            balance_status_grid.addWidget(label, row, 3, ALIGNMENT)
            self.balance_labels_raw_mass_at.append(label)
            label = QLabel("-")
            balance_status_grid.addWidget(label, row, 4, ALIGNMENT)
            self.balance_labels_stable_mass.append(label)
            label = QLabel("-")
            balance_status_grid.addWidget(label, row, 5, ALIGNMENT)
            self.balance_labels_stable_mass_at.append(label)
            label = QLabel("-")
            balance_status_grid.addWidget(label, row, 6, ALIGNMENT)
            self.balance_labels_idmass.append(label)
            label = QLabel("-")
            balance_status_grid.addWidget(label, row, 7, ALIGNMENT)
            self.balance_labels_rfid.append(label)
            label = QLabel("-")
            balance_status_grid.addWidget(label, row, 8, ALIGNMENT)
            self.balance_labels_idmass_at.append(label)

        whisker_status_grid.addWidget(QLabel("<b>Whisker server</b>"),
                                      0, 0, ALIGNMENT)
        whisker_status_grid.addWidget(QLabel("<b>Port</b>"), 0, 1, ALIGNMENT)
        whisker_status_grid.addWidget(QLabel("<b>Status</b>"), 0, 2, ALIGNMENT)
        self.whisker_label_server = QLabel(server_info)
        self.whisker_label_port = QLabel(port_info)
        self.whisker_label_status = QLabel(status_info)
        whisker_status_grid.addWidget(self.whisker_label_server,
                                      1, 0, ALIGNMENT)
        whisker_status_grid.addWidget(self.whisker_label_port,
                                      1, 1, ALIGNMENT)
        whisker_status_grid.addWidget(self.whisker_label_status,
                                      1, 2, ALIGNMENT)

    @pyqtSlot(RfidEvent)
    @exit_on_exception
    def on_rfid(self, rfid_event: RfidEvent) -> None:
        if not isinstance(rfid_event, RfidEvent):
            log.critical("Bad rfid_event: {}".format(rfid_event))
            return
        rfid_index = self.rfid_id_to_idx[rfid_event.reader_id]
        self.rfid_labels_rfid[rfid_index].setText(str(rfid_event.rfid))
        self.rfid_labels_at[rfid_index].setText(
            rfid_event.timestamp.strftime(GUI_TIME_FORMAT))

    @pyqtSlot(MassEvent)
    @exit_on_exception
    def on_mass(self, mass_event: MassEvent) -> None:
        if not isinstance(mass_event, MassEvent):
            log.critical("Bad mass_event: {}".format(mass_event))
            return
        rfid_index = self.rfid_id_to_idx[mass_event.reader_id]
        # For all mass events:
        self.balance_labels_raw_mass[rfid_index].setText(
            GUI_MASS_FORMAT % mass_event.mass_kg)
        self.balance_labels_raw_mass_at[rfid_index].setText(
            mass_event.timestamp.strftime(GUI_TIME_FORMAT))
        # For locked mass events:
        if mass_event.stable:
            self.balance_labels_stable_mass[rfid_index].setText(
                GUI_MASS_FORMAT % mass_event.mass_kg)
            self.balance_labels_stable_mass_at[rfid_index].setText(
                mass_event.timestamp.strftime(GUI_TIME_FORMAT))
        # For locked, identified mass events:
        if mass_event.locked:
            self.balance_labels_idmass[rfid_index].setText(
                GUI_MASS_FORMAT % mass_event.mass_kg)
            self.balance_labels_rfid[rfid_index].setText(str(mass_event.rfid))
            self.balance_labels_idmass_at[rfid_index].setText(
                mass_event.timestamp.strftime(GUI_TIME_FORMAT))

    @pyqtSlot()
    @exit_on_exception
    def on_whisker_state_connected(self) -> None:
        self.whisker_label_status.setText("Connected")

    @exit_on_exception
    def on_whisker_state_disconnected(self) -> None:
        self.whisker_label_status.setText("Disconnected")

    # noinspection PyUnusedLocal
    @pyqtSlot(arrow.Arrow, int)
    @exit_on_exception
    def on_whisker_pingack_received(self, timestamp: arrow.Arrow,
                                    whisker_timestamp: int) -> None:
        self.status("Whisker server acknowledges ping (timestamp {})".format(
            timestamp))

    @pyqtSlot(int, str)
    @exit_on_exception
    def on_rfid_state(self, reader_index: int, state: str) -> None:
        if reader_index < 0 or reader_index >= len(self.rfid_labels_status):
            log.critical("bad reader_index: {}".format(repr(reader_index)))
            return
        self.rfid_labels_status[reader_index].setText(state)

    @pyqtSlot(int, str)
    @exit_on_exception
    def on_balance_state(self, balance_index: int, state: str) -> None:
        if balance_index < 0 or balance_index >= len(self.rfid_labels_status):
            log.critical("bad balance_index: {}".format(repr(balance_index)))
            return
        self.balance_labels_status[balance_index].setText(state)

    def set_button_states(self) -> None:
        running = self.anything_running()
        sqlite = database_is_sqlite(self.dbsettings)
        self.configure_button.setText(
            'View configuration' if running and not sqlite else '&Configure')
        self.configure_button.setEnabled(not running or not sqlite)
        self.start_button.setEnabled(not running)
        self.stop_button.setEnabled(running)
        self.reset_rfids_button.setEnabled(running)
        self.ping_balances_button.setEnabled(running)
        self.calibrate_balances_button.setEnabled(running)
        self.ping_whisker_button.setEnabled(running)


# =============================================================================
# Extra derived classes
# =============================================================================

class RfidKeepCheckListModel(GenericListModel):
    def item_deletable(self, rowindex: int) -> bool:
        return not (
            self.session.query(
                exists().where(RfidEventRecord.reader_id ==
                               self.listdata[rowindex].id)
            ).scalar()
        )


class BalanceKeepCheckListModel(GenericListModel):
    def item_deletable(self, rowindex: int) -> bool:
        return not (
            self.session.query(
                exists().where(MassEventRecord.balance_id ==
                               self.listdata[rowindex].id)
            ).scalar()
        )


# =============================================================================
# Edit main config
# =============================================================================

class MasterConfigWindow(QDialog, TransactionalEditDialogMixin):
    """
    Edits a MasterConfig object.
    """

    # noinspection PyUnresolvedReferences
    def __init__(self,
                 session: Session,
                 config: MasterConfig,
                 parent: QObject = None,
                 readonly: bool = False,
                 **kwargs) -> None:
        main_layout = QVBoxLayout()
        super().__init__(session=session,
                         obj=config,
                         layout=main_layout,
                         readonly=readonly,
                         parent=parent,
                         **kwargs)

        # Title
        self.setWindowTitle("Configure Starfeeder")

        # Elements
        self.rfid_effective_time_edit = QLineEdit()
        self.server_edit = QLineEdit(placeholderText="typically: localhost")
        self.port_edit = QLineEdit(placeholderText="typically: 3233")
        self.wcm_prefix_edit = QLineEdit()
        self.rfid_lv = ModalEditListView(session, RfidConfigDialog,
                                         readonly=readonly)
        self.rfid_lv.selected_maydelete.connect(self.set_rfid_button_states)
        self.balance_lv = ModalEditListView(session, BalanceConfigDialog,
                                            readonly=readonly)
        self.balance_lv.selected_maydelete.connect(
            self.set_balance_button_states)

        # Layout/buttons
        logic_group = StyledQGroupBox('Task logic')
        lform = QFormLayout()
        lform.addRow("RFID effective time (s)<br>This is the time that an RFID"
                     " event ‘persists’ for.", self.rfid_effective_time_edit)
        logic_group.setLayout(lform)

        whisker_group = StyledQGroupBox('Whisker')
        wform = QFormLayout()
        wform.addRow("Whisker server", self.server_edit)
        wform.addRow("Whisker port", self.port_edit)
        wform.addRow("Whisker client message prefix", self.wcm_prefix_edit)
        whisker_group.setLayout(wform)

        rfid_group = StyledQGroupBox('RFID readers')
        rfid_layout_1 = QHBoxLayout()
        rfid_layout_2 = QVBoxLayout()
        if not readonly:
            self.rfid_add_button = QPushButton('Add')
            self.rfid_add_button.clicked.connect(self.add_rfid)
            self.rfid_remove_button = QPushButton('Remove')
            self.rfid_remove_button.clicked.connect(self.remove_rfid)
            rfid_layout_2.addWidget(self.rfid_add_button)
            rfid_layout_2.addWidget(self.rfid_remove_button)
        self.rfid_edit_button = QPushButton('View' if readonly else 'Edit')
        self.rfid_edit_button.clicked.connect(self.edit_rfid)
        # ... or double-click
        rfid_layout_2.addWidget(self.rfid_edit_button)
        rfid_layout_2.addStretch(1)
        rfid_layout_1.addWidget(self.rfid_lv)
        rfid_layout_1.addLayout(rfid_layout_2)
        rfid_group.setLayout(rfid_layout_1)

        balance_group = StyledQGroupBox('Balances')
        balance_layout_1 = QHBoxLayout()
        balance_layout_2 = QVBoxLayout()
        if not readonly:
            self.balance_add_button = QPushButton('Add')
            self.balance_add_button.clicked.connect(self.add_balance)
            self.balance_remove_button = QPushButton('Remove')
            self.balance_remove_button.clicked.connect(self.remove_balance)
            balance_layout_2.addWidget(self.balance_add_button)
            balance_layout_2.addWidget(self.balance_remove_button)
        self.balance_edit_button = QPushButton('View' if readonly else 'Edit')
        self.balance_edit_button.clicked.connect(self.edit_balance)
        balance_layout_2.addWidget(self.balance_edit_button)
        balance_layout_2.addStretch(1)
        balance_layout_1.addWidget(self.balance_lv)
        balance_layout_1.addLayout(balance_layout_2)
        balance_group.setLayout(balance_layout_1)

        main_layout.addWidget(logic_group)
        main_layout.addWidget(whisker_group)
        main_layout.addWidget(rfid_group)
        main_layout.addWidget(balance_group)

        self.set_rfid_button_states(False, False)
        self.set_balance_button_states(False, False)

        # Pass in data
        self.object_to_dialog(self.obj)

    def object_to_dialog(self, obj: MasterConfig) -> None:
        self.rfid_effective_time_edit.setText(str(
            obj.rfid_effective_time_s
            if obj.rfid_effective_time_s is not None else ''))
        self.server_edit.setText(obj.server)
        self.port_edit.setText(str(obj.port or ''))
        self.wcm_prefix_edit.setText(obj.wcm_prefix)
        rfid_lm = RfidKeepCheckListModel(obj.rfidreader_configs,
                                         self.session, self)
        self.rfid_lv.setModel(rfid_lm)
        balance_lm = BalanceKeepCheckListModel(obj.balance_configs,
                                               self.session, self)
        self.balance_lv.setModel(balance_lm)

    def dialog_to_object(self, obj: MasterConfig) -> None:
        # Master config validation and cross-checks.
        # ---------------------------------------------------------------------
        # Basic checks
        # ---------------------------------------------------------------------
        try:
            obj.rfid_effective_time_s = float(
                self.rfid_effective_time_edit.text())
            assert obj.rfid_effective_time_s > 0
        except:
            raise ValidationError("Invalid RFID effective time")
        try:
            obj.server = self.server_edit.text()
            assert len(obj.server) > 0
        except:
            raise ValidationError("Invalid server name")
        try:
            obj.port = int(self.port_edit.text())
            assert obj.port > 0
        except:
            raise ValidationError("Invalid port number")
        # ---------------------------------------------------------------------
        # Duplicate device ports, or names?
        # ---------------------------------------------------------------------
        name_port_pairs = (
            [(r.name, r.port) for r in obj.rfidreader_configs] +
            [(b.name, b.port) for b in obj.balance_configs]
        )
        names = [x[0] for x in name_port_pairs]
        # noinspection PyArgumentList
        duplicate_names = [
            item for item, count in collections.Counter(names).items()
            if count > 1
        ]
        if duplicate_names:
            raise ValidationError(
                "Devices have duplicate names!<br>"
                "Names: {}.".format(duplicate_names))
        ports = [x[1] for x in name_port_pairs]
        if platform.system() == 'Windows':
            # Windows is case-insensitive; e.g. com1, COM1
            ports = [x.upper() for x in ports]
        # noinspection PyArgumentList
        duplicate_ports = [
            item for item, count in collections.Counter(ports).items()
            if count > 1
        ]
        names_of_duplicate_ports = [x[0] for x in name_port_pairs
                                    if x[1] in duplicate_ports]
        if duplicate_ports:
            raise ValidationError(
                "More than one device on a single serial port!<br>"
                "Names: {}.<br>Ports: {}".format(names_of_duplicate_ports,
                                                 duplicate_ports))
        obj.wcm_prefix = self.wcm_prefix_edit.text()
        # ---------------------------------------------------------------------
        # Balances without a paired RFID, or with duplicate pairs?
        # ---------------------------------------------------------------------
        used_reader_names = []
        for balance_config in obj.balance_configs:
            if balance_config.reader is None:
                raise ValidationError(
                    "Balance {} has no paired RFID reader".format(
                        balance_config.name))
            if balance_config.enabled and not balance_config.reader.enabled:
                raise ValidationError(
                    "Balance {} is using RFID reader {},<br>"
                    "but this is disabled".format(
                        balance_config.name,
                        balance_config.reader.name))
            if balance_config.reader.name in used_reader_names:
                raise ValidationError(
                    "More than one balance is trying to use reader {}".format(
                        balance_config.reader.name))
            used_reader_names.append(balance_config.reader.name)

    @pyqtSlot()
    def add_rfid(self) -> None:
        config = RfidReaderConfig(master_config_id=self.obj.id)
        self.rfid_lv.add_in_nested_transaction(config)

    @pyqtSlot()
    def remove_rfid(self) -> None:
        self.rfid_lv.remove_selected()

    @pyqtSlot()
    def edit_rfid(self) -> None:
        self.rfid_lv.edit_selected()

    @pyqtSlot()
    def add_balance(self) -> None:
        config = BalanceConfig(master_config_id=self.obj.id)
        self.balance_lv.add_in_nested_transaction(config)

    @pyqtSlot()
    def remove_balance(self) -> None:
        self.balance_lv.remove_selected()

    @pyqtSlot()
    def edit_balance(self) -> None:
        self.balance_lv.edit_selected()

    @pyqtSlot(bool, bool)
    def set_rfid_button_states(self, selected: bool, maydelete: bool) -> None:
        if not self.readonly:
            self.rfid_remove_button.setEnabled(maydelete)
        self.rfid_edit_button.setEnabled(selected)

    @pyqtSlot(bool, bool)
    def set_balance_button_states(self, selected: bool,
                                  maydelete: bool) -> None:
        if not self.readonly:
            self.balance_remove_button.setEnabled(maydelete)
        self.balance_edit_button.setEnabled(selected)


# =============================================================================
# Dialog components for serial config
# =============================================================================

class SerialPortMixin(object):
    FLOW_NONE = 0
    FLOW_XONXOFF = 1
    FLOW_RTSCTS = 2
    FLOW_DTRDSR = 3

    def __init__(self,
                 port_options: List[str] = None,
                 allow_other_port: bool = True,
                 baudrate_options: List[int] = None,
                 allow_other_baudrate: bool = False,
                 bytesize_options: List[int] = None,
                 parity_options: List[str] = None,
                 stopbits_options: List[float] = None,
                 flow_options: List[int] = None,
                 **kwargs) -> None:
        """
        Always helpful to have allow_other_port=True on Linux, because you can
        create new debugging ports at the drop of a hat, and the serial port
        enumerator may not notice.
        """
        super().__init__(**kwargs)
        self.sp_port_options = port_options
        self.sp_allow_other_port = allow_other_port
        self.sp_baudrate_options = baudrate_options
        self.sp_allow_other_baudrate = allow_other_baudrate

        bytesize_map = [
            (serial.FIVEBITS, "&5"),
            (serial.SIXBITS, "&6"),
            (serial.SEVENBITS, "&7"),
            (serial.EIGHTBITS, "&8"),
        ]
        if bytesize_options:
            bytesize_map = [x for x in bytesize_map
                            if x[0] in bytesize_options]

        parity_map = [
            (serial.PARITY_NONE, "&None"),
            (serial.PARITY_EVEN, "&Even"),
            (serial.PARITY_ODD, "&Odd"),
            (serial.PARITY_MARK, "Mark (rare)"),
            (serial.PARITY_SPACE, "Space (rare)"),
        ]
        if parity_options:
            parity_map = [x for x in parity_map if x[0] in parity_options]

        stopbits_map = [
            (serial.STOPBITS_ONE, "&1"),
            (serial.STOPBITS_ONE_POINT_FIVE, "1.5 (rare)"),
            (serial.STOPBITS_TWO, "&2"),
        ]
        if stopbits_options:
            stopbits_map = [x for x in stopbits_map
                            if x[0] in stopbits_options]

        flow_map = [
            (self.FLOW_NONE, "None (not advised)"),
            (self.FLOW_XONXOFF, "&XON/XOFF software flow control"),
            (self.FLOW_RTSCTS, "&RTS/CTS hardware flow control"),
            (self.FLOW_DTRDSR, "&DTR/DSR hardware flow control"),
        ]
        if flow_options:
            flow_map = [x for x in flow_map if x[0] in flow_options]

        form = QFormLayout()
        if self.sp_port_options:
            self.sp_port_combo = QComboBox()
            self.sp_port_combo.setEditable(allow_other_port)
            self.sp_port_combo.addItems(port_options)
            sp_port_thing = self.sp_port_combo
        else:
            self.sp_port_edit = QLineEdit()
            sp_port_thing = self.sp_port_edit
        form.addRow("Serial port", sp_port_thing)
        if baudrate_options:
            self.sp_baudrate_combo = QComboBox()
            self.sp_baudrate_combo.setEditable(allow_other_baudrate)
            self.sp_baudrate_combo.addItems([str(x) for x in baudrate_options])
            sp_baudrate_thing = self.sp_baudrate_combo
        else:
            self.sp_baudrate_edit = QLineEdit()
            sp_baudrate_thing = self.sp_baudrate_edit
        form.addRow("Speed in bits per second", sp_baudrate_thing)

        sp_bytesize_group = StyledQGroupBox("Data bits")
        self.sp_bytesize_rg = RadioGroup(bytesize_map,
                                         default=serial.EIGHTBITS)
        sp_bytesize_layout = QHBoxLayout()
        self.sp_bytesize_rg.add_buttons_to_layout(sp_bytesize_layout)
        sp_bytesize_layout.addStretch(1)
        sp_bytesize_group.setLayout(sp_bytesize_layout)

        sp_parity_group = StyledQGroupBox("Parity bit")
        self.sp_parity_rg = RadioGroup(parity_map, default=serial.PARITY_NONE)
        sp_parity_layout = QHBoxLayout()
        self.sp_parity_rg.add_buttons_to_layout(sp_parity_layout)
        sp_parity_layout.addStretch(1)
        sp_parity_group.setLayout(sp_parity_layout)

        sp_stop_group = StyledQGroupBox("Stop bits")
        self.sp_stop_rg = RadioGroup(stopbits_map, default=serial.STOPBITS_ONE)
        sp_stop_layout = QHBoxLayout()
        self.sp_stop_rg.add_buttons_to_layout(sp_stop_layout)
        sp_stop_layout.addStretch(1)
        sp_stop_group.setLayout(sp_stop_layout)

        # It's daft to use >1 method of flow control. So use a single radio.
        sp_flow_group = StyledQGroupBox("Flow control")
        # RadioGroup expects         Iterable[Tuple[str, Any]]
        # and we're going to give it     List[Tuple[int, str]]
        # noinspection PyTypeChecker
        self.sp_flow_rg = RadioGroup(flow_map, default=self.FLOW_RTSCTS)
        sp_flow_layout = QVBoxLayout()
        self.sp_flow_rg.add_buttons_to_layout(sp_flow_layout)
        sp_flow_group.setLayout(sp_flow_layout)

        vlayout = QVBoxLayout()
        vlayout.addLayout(form)
        vlayout.addWidget(sp_bytesize_group)
        vlayout.addWidget(sp_parity_group)
        vlayout.addWidget(sp_stop_group)
        vlayout.addWidget(sp_flow_group)

        self.sp_group = StyledQGroupBox('Serial port settings')
        self.sp_group.setLayout(vlayout)

    def serial_port_group_to_object(self, obj: SerialPortConfigMixin) -> None:
        try:
            if self.sp_port_options:
                obj.port = self.sp_port_combo.currentText()
            else:
                obj.port = self.sp_port_edit.text()
            assert len(obj.port) > 0
        except:
            raise ValidationError("Invalid serial port name")
        try:
            if self.sp_baudrate_options:
                obj.baudrate = int(self.sp_baudrate_combo.currentText())
            else:
                obj.baudrate = int(self.sp_baudrate_edit.text())
            assert obj.baudrate > 0
        except:
            raise ValidationError("Invalid speed")
        obj.bytesize = self.sp_bytesize_rg.get_value()
        obj.parity = self.sp_parity_rg.get_value()
        obj.stopbits = self.sp_stop_rg.get_value()
        flow = self.sp_flow_rg.get_value()
        obj.xonxoff = flow == self.FLOW_XONXOFF
        obj.rtscts = flow == self.FLOW_RTSCTS
        obj.dsrdtr = flow == self.FLOW_DTRDSR

    def object_to_serial_port_group(self, obj: SerialPortConfigMixin) -> None:
        if self.sp_port_options:
            if obj.port in self.sp_port_options:
                index = self.sp_port_options.index(obj.port)
                self.sp_port_combo.setCurrentIndex(index)
            elif self.sp_allow_other_port:
                self.sp_port_combo.setEditText(obj.port)
            else:
                self.sp_port_combo.setCurrentIndex(0)
        else:
            self.sp_port_edit.setText(obj.port)
        if self.sp_baudrate_options:
            if obj.baudrate in self.sp_baudrate_options:
                index = self.sp_baudrate_options.index(obj.baudrate)
                self.sp_baudrate_combo.setCurrentIndex(index)
            elif self.sp_allow_other_baudrate:
                self.sp_baudrate_combo.setEditText(str(obj.baudrate))
            else:
                self.sp_baudrate_combo.setCurrentIndex(0)
        else:
            self.sp_baudrate_edit.setText(str(obj.baudrate or ''))
        self.sp_bytesize_rg.set_value(obj.bytesize)
        self.sp_parity_rg.set_value(obj.parity)
        self.sp_stop_rg.set_value(obj.stopbits)
        if obj.rtscts:
            flow = self.FLOW_RTSCTS
        elif obj.dsrdtr:
            flow = self.FLOW_DTRDSR
        elif obj.xonxoff:
            flow = self.FLOW_XONXOFF
        else:
            flow = self.FLOW_NONE
        self.sp_flow_rg.set_value(flow)


# =============================================================================
# Get available serial ports
# =============================================================================
# Do it live, in case they change live.

def get_available_serial_ports() -> List[str]:
    return sorted([item[0] for item in comports()], key=natural_keys)
    # comports() returns a list/tuple of tuples: (port, desc, hwid)


# =============================================================================
# Edit RFID config
# =============================================================================

class RfidConfigDialog(QDialog, TransactionalEditDialogMixin,
                       SerialPortMixin):

    def __init__(self,
                 session: Session,
                 rfid_config: RfidReaderConfig,
                 parent: QObject = None,
                 readonly: bool = False,
                 **kwargs):
        top_layout = QVBoxLayout()
        super().__init__(
            # for SerialPortMixin       [3]
            port_options=get_available_serial_ports(),
            baudrate_options=[9600],
            bytesize_options=[serial.EIGHTBITS],
            parity_options=[serial.PARITY_NONE],
            stopbits_options=[serial.STOPBITS_ONE],
            # for TransactionalEditDialogMixin
            session=session,
            obj=rfid_config,
            layout=top_layout,
            readonly=readonly,
            # for QDialog
            parent=parent,
            # Anyone else?
            **kwargs
        )

        # Title
        self.setWindowTitle("Configure RFID reader")

        # Elements
        self.enabled_group = StyledQGroupBox("Enabled")
        self.enabled_group.setCheckable(True)
        self.id_value_label = QLabel()
        self.name_edit = QLineEdit()
        warning1 = QLabel(RENAME_WARNING)
        warning2 = QLabel("<b>NOTE:</b> the intended RFID devices are fixed "
                          "in hardware to 9600 bps, 8N1</b>")  # [3]

        # Layout
        form = QFormLayout()
        form.addRow(DEVICE_ID_LABEL, self.id_value_label)
        form.addRow("RFID name", self.name_edit)

        main_layout = QVBoxLayout()
        main_layout.addLayout(form)
        main_layout.addWidget(warning1)
        main_layout.addWidget(warning2)
        main_layout.addWidget(self.sp_group)

        self.enabled_group.setLayout(main_layout)
        top_layout.addWidget(self.enabled_group)

        # Pass in data
        self.object_to_dialog(self.obj)

    def object_to_dialog(self, obj: RfidReaderConfig) -> None:
        self.enabled_group.setChecked(obj.enabled)
        self.id_value_label.setText(str(obj.id))
        self.name_edit.setText(obj.name)
        self.object_to_serial_port_group(obj)

    def dialog_to_object(self, obj: RfidReaderConfig) -> None:
        obj.enabled = self.enabled_group.isChecked()
        try:
            obj.name = self.name_edit.text()
            assert len(obj.name) > 0
        except:
            raise ValidationError("Invalid name")
        self.serial_port_group_to_object(obj)


# =============================================================================
# Edit balance config
# =============================================================================

class BalanceConfigDialog(QDialog, TransactionalEditDialogMixin,
                          SerialPortMixin):

    def __init__(self,
                 session: Session,
                 balance_config: BalanceConfig,
                 parent: QObject = None,
                 readonly: bool = False,
                 **kwargs):
        top_layout = QVBoxLayout()
        super().__init__(
            # for SerialPortMixin   [4]
            port_options=get_available_serial_ports(),
            baudrate_options=[1200, 2400, 4800, 9600, 19200, 38400],
            bytesize_options=[serial.EIGHTBITS],
            parity_options=[serial.PARITY_NONE, serial.PARITY_EVEN],
            stopbits_options=[serial.STOPBITS_ONE],
            flow_options=[SerialPortMixin.FLOW_NONE,
                          SerialPortMixin.FLOW_XONXOFF],
            # for TransactionalEditDialogMixin
            session=session,
            obj=balance_config,
            layout=top_layout,
            readonly=readonly,
            # for QDialog
            parent=parent,
            # Anyone else?
            **kwargs
        )
        # RTS/CTS sometimes seems to break it.
        # Manual mentions XON/XOFF only (p15), and says that its serial
        # interface is RS-485, 2-wire, half-duplex (p4, 5).

        reader_map = []
        readers = (
            session.query(RfidReaderConfig)
            .filter(RfidReaderConfig.enabled == True)  # http://stackoverflow.com/questions/18998010  # noqa
            .all()
        )
        for reader in readers:
            reader_map.append((reader.id, reader.name))
        if reader_map:
            reader_map.sort(key=lambda x: natural_keys(x[1]))
            self.reader_ids, self.reader_names = zip(*reader_map)
        else:
            self.reader_ids = []
            self.reader_names = []

        self.setWindowTitle("Configure balance")

        warning1 = QLabel(RENAME_WARNING)
        warning2 = QLabel(
            "<b>NOTE:</b> the intended balance devices default to 9600 bps, "
            "8E1,<br>and are restricted in their serial options")  # [4]
        self.enabled_group = StyledQGroupBox("Enabled")
        self.enabled_group.setCheckable(True)
        self.id_value_label = QLabel()
        self.name_edit = QLineEdit()
        self.reader_combo = QComboBox()
        self.reader_combo.addItems(self.reader_names)
        self.reader_combo.setEditable(False)
        self.asf_combo = QComboBox()
        self.asf_combo.addItems(list(str(x) for x in POSSIBLE_ASF_MODES))
        self.asf_combo.setEditable(False)
        self.fast_filter_check = QCheckBox()
        self.measurement_rate_hz_combo = QComboBox()
        self.measurement_rate_hz_combo.addItems(
            [str(x) for x in POSSIBLE_RATES_HZ])
        self.measurement_rate_hz_combo.setEditable(False)
        self.stability_n_edit = QLineEdit()
        self.tolerance_kg_edit = QLineEdit()
        self.min_mass_kg_edit = QLineEdit()
        self.unlock_mass_kg_edit = QLineEdit()
        self.refload_mass_kg_edit = QLineEdit()
        self.zero_value_label = QLabel()
        self.refload_value_label = QLabel()
        self.read_continuously_check = QCheckBox()

        form1 = QFormLayout()
        form1.addRow(DEVICE_ID_LABEL, self.id_value_label)
        form1.addRow("Balance name", self.name_edit)
        form1.addRow("Paired RFID reader", self.reader_combo)

        meas_group = StyledQGroupBox('Measurement settings')
        form2 = QFormLayout()
        form2.addRow("Amplifier signal filter (ASF) mode (0 = none; "
                     "see p37 of manual)", self.asf_combo)
        form2.addRow("Fast response filter (FMD; see p37 of manual)",
                     self.fast_filter_check)
        form2.addRow("Measurement rate (Hz)", self.measurement_rate_hz_combo)
        form2.addRow("Number of consecutive readings judged for stability",
                     self.stability_n_edit)
        form2.addRow("Stability tolerance (kg) (range [max - min] of<br>"
                     "consecutive readings must not exceed this)",
                     self.tolerance_kg_edit)
        form2.addRow("Minimum mass for detection (kg)", self.min_mass_kg_edit)
        form2.addRow("Mass below which balance will unlock (kg)",
                     self.unlock_mass_kg_edit)
        form2.addRow("Reference (calibration) mass (kg)",
                     self.refload_mass_kg_edit)
        form2.addRow("Zero (tare) calibration point", self.zero_value_label)
        form2.addRow("Reference mass calibration point",
                     self.refload_value_label)
        form2.addRow("Read continuously (inefficient)",
                     self.read_continuously_check)

        mg_vl = QVBoxLayout()
        mg_vl.addLayout(form2)
        meas_group.setLayout(mg_vl)

        main_layout = QVBoxLayout()
        main_layout.addWidget(warning1)
        main_layout.addLayout(form1)
        main_layout.addWidget(meas_group)
        main_layout.addWidget(warning2)
        main_layout.addWidget(self.sp_group)

        self.enabled_group.setLayout(main_layout)
        top_layout.addWidget(self.enabled_group)

        # Pass in data
        self.object_to_dialog(self.obj)

    def object_to_dialog(self, obj: BalanceConfig) -> None:
        self.enabled_group.setChecked(obj.enabled or False)
        self.id_value_label.setText(str(obj.id))
        self.name_edit.setText(obj.name)
        if obj.reader_id in self.reader_ids:
            self.reader_combo.setCurrentIndex(
                self.reader_ids.index(obj.reader_id))
        else:
            self.reader_combo.setCurrentIndex(0)
        if obj.measurement_rate_hz in POSSIBLE_RATES_HZ:
            self.measurement_rate_hz_combo.setCurrentIndex(
                POSSIBLE_RATES_HZ.index(obj.measurement_rate_hz))
        if obj.amp_signal_filter_mode in POSSIBLE_ASF_MODES:
            self.asf_combo.setCurrentIndex(
                POSSIBLE_ASF_MODES.index(obj.amp_signal_filter_mode))
        self.fast_filter_check.setChecked(obj.fast_response_filter or False)
        self.stability_n_edit.setText(str(obj.stability_n))
        self.tolerance_kg_edit.setText(str(obj.tolerance_kg))
        self.min_mass_kg_edit.setText(str(obj.min_mass_kg))
        self.unlock_mass_kg_edit.setText(str(obj.unlock_mass_kg))
        self.refload_mass_kg_edit.setText(str(obj.refload_mass_kg))
        self.zero_value_label.setText(str(obj.zero_value))
        self.refload_value_label.setText(str(obj.refload_value))
        self.read_continuously_check.setChecked(obj.read_continuously or False)
        self.object_to_serial_port_group(obj)

    def dialog_to_object(self, obj: BalanceConfig) -> None:
        obj.enabled = self.enabled_group.isChecked()
        try:
            obj.name = self.name_edit.text()
            assert len(obj.name) > 0
        except:
            raise ValidationError("Invalid name")
        reader_name = self.reader_combo.currentText()
        try:
            reader_index = self.reader_names.index(reader_name)
            obj.reader_id = self.reader_ids[reader_index]
        except:
            raise ValidationError("Invalid reader")
        try:
            obj.measurement_rate_hz = int(
                self.measurement_rate_hz_combo.currentText())
            assert obj.measurement_rate_hz in POSSIBLE_RATES_HZ
        except:
            raise ValidationError("Invalid measurement_rate_hz")
        try:
            obj.amp_signal_filter_mode = int(self.asf_combo.currentText())
            assert obj.amp_signal_filter_mode in POSSIBLE_ASF_MODES
        except:
            raise ValidationError("Invalid amp_signal_filter_mode")
        obj.fast_response_filter = self.fast_filter_check.isChecked()
        try:
            obj.stability_n = int(self.stability_n_edit.text())
            assert obj.stability_n > 1
        except:
            raise ValidationError("Invalid stability_n")
        try:
            obj.tolerance_kg = float(self.tolerance_kg_edit.text())
            assert obj.tolerance_kg > 0
        except:
            raise ValidationError("Invalid tolerance_kg")
        try:
            obj.min_mass_kg = float(self.min_mass_kg_edit.text())
            assert obj.min_mass_kg > 0
        except:
            raise ValidationError("Invalid min_mass_kg")
        try:
            obj.unlock_mass_kg = float(self.unlock_mass_kg_edit.text())
            assert obj.unlock_mass_kg > 0
        except:
            raise ValidationError("Invalid unlock_mass_kg")
        try:
            obj.refload_mass_kg = float(self.refload_mass_kg_edit.text())
            assert obj.refload_mass_kg > 0
        except:
            raise ValidationError("Invalid refload_mass_kg")
        obj.read_continuously = self.read_continuously_check.isChecked()
        self.serial_port_group_to_object(obj)
        if obj.unlock_mass_kg >= obj.min_mass_kg:
            raise ValidationError(
                "unlock_mass_kg must be less than min_mass_kg")


# =============================================================================
# Tare/calibrate balances
# =============================================================================

class CalibrateBalancesWindow(QDialog):
    # noinspection PyUnresolvedReferences
    def __init__(self, balance_owners: List[BalanceOwner],
                 parent: QObject = None, **kwargs) -> None:
        super().__init__(parent=parent, **kwargs)
        self.setWindowTitle("Calibrate balances")

        grid = QGridLayout()
        for i, balance in enumerate(balance_owners):
            grid.addWidget(QLabel("Balance {}:".format(
                balance.balance_id, balance.name)), i, 0)
            tare_button = QPushButton("&Tare (zero)")
            tare_button.clicked.connect(balance.tare)
            grid.addWidget(tare_button, i, 1)
            calibrate_button = QPushButton("&Calibrate to {} kg".format(
                balance.refload_mass_kg))
            calibrate_button.clicked.connect(balance.calibrate)
            grid.addWidget(calibrate_button, i, 2)

        ok_buttons = QDialogButtonBox(QDialogButtonBox.Ok,
                                      Qt.Horizontal, self)
        ok_buttons.accepted.connect(self.accept)

        vlayout = QVBoxLayout(self)
        vlayout.addLayout(grid)
        vlayout.addWidget(ok_buttons)



#!/usr/bin/env python
# starfeeder/rfid.py

"""
    Copyright (C) 2015-2015 Rudolf Cardinal (rudolf@pobox.com).

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

import re
from typing import Optional

import arrow
import bitstring
from PyQt5.QtCore import QObject, QTimer, pyqtSignal, pyqtSlot
from whisker.qt import exit_on_exception

from starfeeder.serial_controller import (
    # CR,
    CRLF,
    # LF,
    LF_STR,
    NO_BYTES,
    SerialController,
    SerialOwner,
)
from starfeeder.models import (
    BalanceConfig,  # for type hints
    RfidEvent,
    RfidReaderConfig,  # for type hints
)

CMD_RESET_1 = "x"  # response: "MULTITAG-125 01" (+/- "S" as a separate line)
CMD_RESET_2 = "z"  # response: "MULTITAG-125 01"
CMD_REQUEST_VERSION = "v"  # response: "MULTITAG-125 01"
CMD_READING_CONTINUES = "c"  # response: nothing -> RFIDs; "S" after next char
CMD_NO_OP_CANCEL = LF_STR
CMD_LOGIN = "l"  # but response: "?"
CMD_READ_PAGE = "r"  # but response: "?"
CMD_WRITE_PAGE = "w"  # but response: "?"
CMD_DENOMINATION_OF_LED = "d"
CMD_ANTENNA_POWER_OFF = "p"  # response: "P"

RESPONSE_COMMAND_INVALID = "?"
RESPONSE_COMMAND_NOT_EXECUTED = "N"
RESPONSE_CONTINUOUS_READ_STOPPED = "S"
RESPONSE_ANTENNA_OFF = "P"

HELLO_REGEX = re.compile(r"^MULTITAG(.*)$")

RESET_PAUSE_MS = 200


def ztag_to_rfid_number(ztag: str) -> Optional[int]:
    """
    Parses a Z-tag from the RFID reader. Examples:
        Z5A2080A70C2C0001 [= RFID 208210000479322; 208 = Denmark]
        Z1FC68BAD50870001 [= RFID 900046000071672; 900 = user-assigned]

    References:
    - [6], though very hard to read; ignored
    - http://www.priority1design.com.au/fdx-b_animal_identification_protocol.html  # noqa
    - https://en.wikipedia.org/wiki/ISO_11784_%26_11785
    - http://www.gizmolab.co.za/fdx-b-protocol/
    - http://python-bitstring.googlecode.com/svn/trunk/doc/bitstring_manual.pdf

    The RFID tag reader uses the FDX-B protocol [6].

    Country codes:
    - https://en.wikipedia.org/wiki/ISO_3166-1
    - UK is 826 decimal

    What I think is going on:
    - The FDX-B protocol sends 128 bits of data:
        11 header bits
        64 identification bits with [plus] 8 control bits
        16 bits of CRC with 2 control bits
        24 bits of extended data with [plus] 3 control bits
        ... 11 + 64 + 8 + 16 + 2 + 24 + 3 = 128
    - The Z tag contains 16 hex characters = 8 bytes = 64 bits.
      So that's probably the 64 identification bits, with the control bits
      stripped out. (Yes; confirmed by match to bar code at the end.)
    - The 64 identification bits look like this:
            (https://en.wikipedia.org/wiki/ISO_11784_%26_11785#FDX)
        38 bits of identification code
            ... NOTE: little-endian (LSB to "left")
        10 bits of country code (or 999 decimal = test transponder)
            ... NOTE: little-endian (LSB to "left")
         1 bit of "additional data block yes (1) or no (0)"
         7 bits of padding
         7 bits of padding
         1 bit of "animal identification yes (1) or no (0)"

    - Reading: bitstring is great. But it can't read a little-endian integer
      directly from a segment that's not a multiple of 8 bytes (not sure why!).
      So we can't map "uintle:38" to national_id; instead, we map "bits:38" to
      national_id_bits, reverse it, and take its number.

    - At the end, rfid_string and rfid_int should match the number on the tag's
      bar code.
    - The final number is at most 15 digits (3 country, 12 national_id).
    - A 32-bit signed integer goes up to 2,147,483,647 -- not enough.
    - A 32-bit unsigned integer goes up to 4,294,967,295 -- not enough.
    - A 64-bit unsigned integer goes up to 18,446,744,073,709,551,615 -- OK.
    - A 64-bit signed integer goes up to +9,223,372,036,854,775,807 -- OK.
    - So as long as it's 64-bit, we're OK.
    """
    if not isinstance(ztag, str) or len(ztag) != 17 or ztag[0] != 'Z':
        return None
    hexcode = ztag[1:]
    bits = bitstring.BitStream(hex=hexcode)
    national_id_bits, country_bits, has_additional_data, is_animal_id = (
        bits.readlist('bits:38, bits:10, bool, pad:7, pad:7, bool'))
    national_id_bits.reverse()  # little-endian to big-endian
    national_id = national_id_bits.uint
    country_bits.reverse()
    country = country_bits.uint
    rfid_string = str(country) + str(national_id).zfill(12)
    rfid_int = int(rfid_string)
    return rfid_int


class RfidController(SerialController):
    """
    Serial controller that knows about RFID reader devices.

    - RFID protocol uses a CR+LF delimiter [3] from device to computer.
    - DO NOT use a terminator from computer to RFID; send single characters
      only. (For example, if you send 'c\n', the 'c' will start the reader
      and the '\n' wil stop it.)
    - Commands/responses from [3].
    """
    rfid_received = pyqtSignal(RfidEvent)

    def __init__(self, reader_config: RfidReaderConfig,
                 balance_config: BalanceConfig, **kwargs) -> None:
        super().__init__(**kwargs)
        self.reader_config = reader_config
        self.balance_config = balance_config
        self.swallow_next_stopped_read = False
        self.reset_timer = QTimer(self)
        # noinspection PyUnresolvedReferences
        self.reset_timer.timeout.connect(self.reset_2)

    def send(self, data: str, delay_ms: int = 0) -> None:
        self.send_str(data, delay_ms)

    @pyqtSlot()
    @exit_on_exception
    def on_start(self) -> None:
        self.reset()

    @pyqtSlot()
    @exit_on_exception
    def on_stop(self) -> None:
        self.send(CMD_NO_OP_CANCEL)  # something to cancel any ongoing read
        self.finished.emit()
        # Inelegant! Risk the writer thread will be terminated before it
        # sends this command. Still, ho-hum.

    @pyqtSlot()
    @exit_on_exception
    def reset(self) -> None:
        self.info("Resetting RFID: phase 1")
        self.swallow_next_stopped_read = True
        self.send(CMD_NO_OP_CANCEL)  # something to cancel any ongoing read
        self.debug("RFID resetting: waiting {} ms for reset".format(
            RESET_PAUSE_MS))
        self.reset_timer.setSingleShot(True)
        self.reset_timer.start(RESET_PAUSE_MS)
        # POTENTIALLY SUBOPTIMAL: these pauses are in the GUI (main) thread,
        # so could be disrupted by something going wrong on the main event
        # loop. The solution would be to use a class derived from SerialWriter
        # and pass it to the SerialController constructor. And then think about
        # how we then see that class instance to connect up its signals.
        # Anyway, let's not over-complicate it for now; resetting isn't a
        # common thing, and the user will have triggered it directly.

    @pyqtSlot()
    def reset_2(self) -> None:
        self.info("Resetting RFID: phase 2")
        self.swallow_next_stopped_read = False
        self.send(CMD_RESET_1)  # again, to make sure we actually reset
        # Wait for the RFID to say hello before we start reading, or the read
        # command will be swallowed up during the reset.

    def start_reading(self) -> None:
        self.info("Asking RFID to start reading")
        self.send(CMD_READING_CONTINUES)

    @pyqtSlot(str, arrow.Arrow)
    def on_receive(self, data: str, timestamp: arrow.Arrow) -> None:
        if not isinstance(data, str):
            self.critical("bad data: {}".format(repr(data)))
            return
        if not isinstance(timestamp, arrow.Arrow):
            self.critical("bad timestamp: {}".format(repr(timestamp)))
            return
        self.debug("Receiving at {}: {}".format(timestamp, repr(data)))
        if data == RESPONSE_COMMAND_INVALID:
            self.debug("RESPONSE_COMMAND_INVALID")
            # We might get this because we send CMD_NO_OP_CANCEL either in the
            # "waiting for command" state (which will produce '?') or in the
            # "continuous read" state, which will cancel the read.
        elif data == RESPONSE_COMMAND_NOT_EXECUTED:
            self.error("RESPONSE_COMMAND_NOT_EXECUTED")
        elif data == RESPONSE_CONTINUOUS_READ_STOPPED:
            self.info("RESPONSE_CONTINUOUS_READ_STOPPED")
            if self.swallow_next_stopped_read:
                self.swallow_next_stopped_read = False
            else:
                self.start_reading()
        elif data == RESPONSE_ANTENNA_OFF:
            self.info("RESPONSE_ANTENNA_OFF")
        elif HELLO_REGEX.match(data):
            self.status("RFID reader says hello: {}".format(repr(data)))
            self.start_reading()
        else:
            # Should be a Z-tag; see above.
            rfid_number = ztag_to_rfid_number(data)
            if rfid_number is None:
                self.error("Received unknown data: {}".format(repr(data)))
                return
            self.debug("rfid number = {}".format(rfid_number))
            # WATCH OUT. Signal "int" values are 32-bit. So we should
            # emit a Python object instead.
            rfid_event = RfidEvent(
                reader_id=self.reader_config.id,
                reader_name=self.reader_config.name,
                rfid=rfid_number,
                timestamp=timestamp)
            self.rfid_received.emit(rfid_event)


class RfidOwner(SerialOwner):  # GUI thread
    # Outwards, to world:
    rfid_received = pyqtSignal(RfidEvent)
    # Inwards, to posessions:
    reset_requested = pyqtSignal()

    def __init__(self, rfid_config: RfidReaderConfig, callback_id: int,
                 parent: QObject = None) -> None:
        # Do not keep a copy of rfid_config; it will expire.
        super().__init__(
            serial_args=rfid_config.get_serial_args(),
            parent=parent,
            callback_id=callback_id,
            name=rfid_config.name,
            rx_eol=CRLF,
            tx_eol=NO_BYTES,
            controller_class=RfidController,
            controller_kwargs=dict(
                reader_config=rfid_config.get_attrdict(),
                balance_config=(rfid_config.balance.get_attrdict()
                                if rfid_config.balance else None),
            ))
        self.reader_id = rfid_config.id  # used by main GUI
        self.name = rfid_config.name  # used by main GUI
        self.reset_requested.connect(self.controller.reset)
        self.controller.rfid_received.connect(self.rfid_received)  # different thread  # noqa

    def reset(self) -> None:
        self.reset_requested.emit()



#!/usr/bin/env python
# starfeeder/test_mimic_rfid_reader.py

"""
Very simple program to mimic an RFID reader talking to a serial port, for
testing Starfeeder. This does not use Qt.
"""

import argparse
import logging

from whisker.logging import configure_logger_for_colour

from starfeeder.constants import (
    LOG_FORMAT,
    LOG_DATEFMT,
)
from starfeeder.common_serial_testing import (
    add_serial_port_args,
    coin,
    CommandLineSerialProcessor,
)
from starfeeder.getch import getch, kbhit
from starfeeder.rfid import (
    CMD_RESET_1,
    CMD_RESET_2,
    CMD_REQUEST_VERSION,
    CMD_READING_CONTINUES,
    CMD_NO_OP_CANCEL,
    CMD_LOGIN,
    CMD_READ_PAGE,
    CMD_WRITE_PAGE,
    CMD_DENOMINATION_OF_LED,
    CMD_ANTENNA_POWER_OFF,

    RESPONSE_COMMAND_INVALID,
    # RESPONSE_COMMAND_NOT_EXECUTED,
    RESPONSE_CONTINUOUS_READ_STOPPED,
    RESPONSE_ANTENNA_OFF,
)
from starfeeder.serial_controller import (
    CRLF,
    # NO_BYTES,
    READ_TIMEOUT_SEC,
)

log = logging.getLogger(__name__)


# =============================================================================
# Behave like an RFID reader
# =============================================================================

class RfidMimic(CommandLineSerialProcessor):
    RFID_ZTAG_EXAMPLES = [
        None,
        'Z5A2080A70C2C0001',
        'Z1FC68BAD50870001',
    ]
    VERSION = "MULTITAG-125 01"

    def __init__(self, args: argparse.Namespace, **kwargs) -> None:
        self.read_timeout_sec = READ_TIMEOUT_SEC
        super().__init__(args,
                         inbound_eol=None, inbound_bytewise=True,
                         outbound_eol=CRLF,
                         read_timeout_sec=self.read_timeout_sec,
                         **kwargs)
        # Device to computer (from our perspective here, outbound): CR+LF
        # Computer to device (here, inbound): no delimiter, single-char commands

        # noinspection PyUnresolvedReferences
        self.mean_time_between_rfids_s = args.mean_time_between_rfids_s
        mean_num_reads_per_rfid = (
            self.mean_time_between_rfids_s / self.read_timeout_sec)
        self.p_send = 1 / mean_num_reads_per_rfid
        self.reading = False
        self.current_rfid = None

        log.info("Use keys 0-{} to select an RFID".format(
            len(self.RFID_ZTAG_EXAMPLES) - 1))

    def start(self) -> None:
        pass

    def spontaneous(self) -> None:
        if kbhit():
            c = getch()
            rfid_index = ord(c) - ord('0')
            if 0 <= rfid_index < len(self.RFID_ZTAG_EXAMPLES):
                self.current_rfid = self.RFID_ZTAG_EXAMPLES[rfid_index]
                log.info("Setting RFID to {}".format(self.current_rfid))
            else:
                log.warning("Key {} invalid; use 0-{}".format(
                    repr(c), len(self.RFID_ZTAG_EXAMPLES) - 1))

        if self.reading:
            # log.debug("shall we send one? p = {}".format(self.p_send))
            if coin(self.p_send):
                if self.current_rfid is not None:
                    self.send_line(self.current_rfid)

    def line_received(self, line: bytes) -> None:
        log.info("Command received: {}".format(repr(line)))

        if line == CMD_RESET_1:
            log.info("CMD_RESET_1 -> send version")
            self.send_line(self.VERSION)

        elif line == CMD_RESET_2:
            log.info("CMD_RESET_2 -> send version")
            self.send_line(self.VERSION)

        elif line == CMD_REQUEST_VERSION:
            log.info("CMD_REQUEST_VERSION -> send version")
            self.send_line(self.VERSION)

        elif line == CMD_READING_CONTINUES:
            log.info("CMD_READING_CONTINUES -> start reading")
            self.start_reading()
            # but no overt reply

        elif line == CMD_NO_OP_CANCEL:
            log.info("CMD_NO_OP_CANCEL -> don't know")

        elif line == CMD_LOGIN:
            log.info("CMD_LOGIN -> RESPONSE_COMMAND_INVALID")
            self.send_line(RESPONSE_COMMAND_INVALID)

        elif line == CMD_READ_PAGE:
            log.info("CMD_READ_PAGE -> RESPONSE_COMMAND_INVALID")
            self.send_line(RESPONSE_COMMAND_INVALID)

        elif line == CMD_WRITE_PAGE:
            log.info("CMD_WRITE_PAGE -> RESPONSE_COMMAND_INVALID")
            self.send_line(RESPONSE_COMMAND_INVALID)

        elif line == CMD_DENOMINATION_OF_LED:
            log.info("CMD_DENOMINATION_OF_LED -> don't know")

        elif line == CMD_ANTENNA_POWER_OFF:
            if self.reading:
                log.info("CMD_ANTENNA_POWER_OFF -> was reading -> "
                         "stop reading; RESPONSE_CONTINUOUS_READ_STOPPED")
                self.send_line(RESPONSE_CONTINUOUS_READ_STOPPED)
                self.stop_reading()
            else:
                log.info("CMD_ANTENNA_POWER_OFF -> wasn't reading -> "
                         "RESPONSE_ANTENNA_OFF (?)")
                self.send_line(RESPONSE_ANTENNA_OFF)  # I think?

        else:
            log.warning("Unknown command received: {} -> "
                        "RESPONSE_COMMAND_INVALID".format(repr(line)))
            self.send_line(RESPONSE_COMMAND_INVALID)

    def start_reading(self) -> None:
        log.info("Starting to pretend to read RFID tags")
        self.reading = True

    def stop_reading(self) -> None:
        log.info("Stopping RFID tag reading")
        self.reading = False


# =============================================================================
# Command-line framework
# =============================================================================

def main() -> None:
    # -------------------------------------------------------------------------
    # Arguments
    # -------------------------------------------------------------------------
    parser = argparse.ArgumentParser(
        description="test_mimic_rfid_reader -- to test Starfeeder",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument(
        '--verbose', '-v', action='store_true', help="Be verbose")
    parser.add_argument(
        '--mean_time_between_rfids_s', type=float, default=1.0,
        help="Mean time between RFID generation (s) when reading")
    add_serial_port_args(parser)
    args = parser.parse_args()
    # -------------------------------------------------------------------------
    # Logging
    # -------------------------------------------------------------------------
    loglevel = logging.DEBUG if args.verbose else logging.INFO
    logging.basicConfig(format=LOG_FORMAT, datefmt=LOG_DATEFMT,
                        level=loglevel)
    rootlogger = logging.getLogger()
    rootlogger.setLevel(loglevel)
    configure_logger_for_colour(rootlogger)  # configure root logger
    # -------------------------------------------------------------------------
    # Go
    # -------------------------------------------------------------------------
    log.info("test_mimic_rfid_reader")
    log.info("args: {}".format(args))

    mimic = RfidMimic(args)
    mimic.run()


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# starfeeder/test_mimic_balance.py

"""
Very simple program to mimic a weighing balance talking to a serial port, for
testing Starfeeder. This does not use Qt.
"""

import argparse
import logging
import random
from time import sleep
from typing import List

from whisker.logging import configure_logger_for_colour

from starfeeder.balance import (
    BAUDRATE_PAUSE_MS,

    CMD_ASCII_RESULT_OUTPUT,
    CMD_DATA_DELIMITER_COMMA_CR_LF,
    CMD_DEACTIVATE_OUTPUT_SCALING,
    CMD_FILTER_TYPE,
    CMD_MEASUREMENT_RATE,
    CMD_NO_OP,
    CMD_QUERY_BAUD_RATE,
    CMD_QUERY_IDENTIFICATION,
    CMD_QUERY_MEASURE,
    CMD_QUERY_OUTPUT_SCALING,
    CMD_QUERY_STATUS,
    CMD_SET_BAUD_RATE,
    CMD_SET_FILTER,
    CMD_STOP_MEASURING,
    CMD_TARE,
    CMD_WARM_RESTART,

    RATE_MAP_HZ_TO_CODE,

    RESPONSE_NONSPECIFIC_OK,
    RESPONSE_UNKNOWN,

    COMMAND_SEPARATOR,
)
from starfeeder.constants import (
    LOG_FORMAT,
    LOG_DATEFMT,
)
from starfeeder.common_serial_testing import (
    add_serial_port_args,
    CommandLineSerialProcessor,
)
from starfeeder.getch import getch, kbhit
from starfeeder.serial_controller import (
    CRLF,
    LF,
    READ_TIMEOUT_SEC,
)

log = logging.getLogger(__name__)


# =============================================================================
# Behave like a balance
# =============================================================================

class OutputMessage(object):
    def __init__(self, delay_s: float, is_measurement: bool,
                 text: str) -> None:
        self.delay_s = delay_s
        self.is_measurement = is_measurement
        self.text = text


class BalanceMimic(CommandLineSerialProcessor):
    MASS_KG_MEANS = [0, 0.1, 0.350, 0.450]
    # need 0 for tare and 0.1 for calibrate
    MASS_KG_SD = 0.0005
    RESTART_TIME_S = 3
    REPLY_STATUS = "000"
    DEFAULT_MAX_VALUE = 100000
    INTEGER_UNITS_PER_KG = 1000
    IDENTIFICATION = "HBM,PW4 MCR2 2kg   ,70820   ,P52"
    # ... from screenshot (number of spaces may be a bit wrong, though)

    def __init__(self, args: argparse.Namespace, **kwargs) -> None:
        self.read_timeout_sec = READ_TIMEOUT_SEC

        super().__init__(args,
                         inbound_eol=[COMMAND_SEPARATOR, CRLF, LF],
                         outbound_eol=CRLF,
                         read_timeout_sec=self.read_timeout_sec,
                         **kwargs)
        # Device to computer (from our perspective here, outbound): CR+LF
        # Computer to device (here, inbound): no delimiter, single-char commands

        self.baudrate = args.baudrate
        self.parity = args.parity
        self.internal_period_s = self.read_timeout_sec
        self.read_frequency_hz = 6
        self.read_period_s = 1 / self.read_frequency_hz
        self.time_to_next_read = 0
        self.tare_mass_kg = 0.00
        self.current_true_mass_kg = self.MASS_KG_MEANS[0]
        self.output_queue = []  # type: List[OutputMessage]

        log.info("Use keys 0-{} to select a mass".format(
            len(self.MASS_KG_MEANS) - 1))

    def start(self) -> None:
        pass

    def spontaneous(self) -> None:
        if kbhit():
            c = getch()
            mass_index = ord(c) - ord('0')
            if 0 <= mass_index < len(self.MASS_KG_MEANS):
                self.current_true_mass_kg = self.MASS_KG_MEANS[mass_index]
                log.info("Setting mass to {} kg (takes effect at next read "
                         "cycle)".format(self.current_true_mass_kg))
            else:
                log.warning("Key {} invalid; use 0-{}".format(
                    repr(c), len(self.MASS_KG_MEANS) - 1))

        while self.output_queue and self.output_queue[0].delay_s <= 0:
            self.send_line(self.output_queue[0].text)
            self.output_queue.pop(0)
        if self.output_queue:
            self.output_queue[0].delay_s -= self.internal_period_s

    def reply(self, text: str) -> None:
        om = OutputMessage(0, False, text)
        self.output_queue.append(om)

    def clear_queue(self) -> None:
        self.output_queue.clear()

    def line_received(self, line: bytes) -> None:
        log.info("Command received: {}".format(repr(line)))
        if line == CMD_ASCII_RESULT_OUTPUT:
            log.info("CMD_ASCII_RESULT_OUTPUT -> RESPONSE_NONSPECIFIC_OK")
            self.reply(RESPONSE_NONSPECIFIC_OK)

        elif line == CMD_DATA_DELIMITER_COMMA_CR_LF:
            log.info("CMD_DATA_DELIMITER_COMMA_CR_LF -> "
                     "RESPONSE_NONSPECIFIC_OK")
            self.reply(RESPONSE_NONSPECIFIC_OK)

        elif line == CMD_DEACTIVATE_OUTPUT_SCALING:
            log.info("CMD_DEACTIVATE_OUTPUT_SCALING -> "
                     "RESPONSE_NONSPECIFIC_OK (?)")
            self.reply(RESPONSE_NONSPECIFIC_OK)

        elif line.startswith(CMD_FILTER_TYPE):
            rest = line[len(CMD_FILTER_TYPE):]
            log.info("CMD_FILTER_TYPE {} -> RESPONSE_NONSPECIFIC_OK".format(
                rest))
            self.reply(RESPONSE_NONSPECIFIC_OK)

        elif line.startswith(CMD_MEASUREMENT_RATE):
            rest = line[len(CMD_MEASUREMENT_RATE):]
            try:
                code = int(rest)
                hz = None
                for lookup_hz, lookup_code in RATE_MAP_HZ_TO_CODE.items():
                    if code == lookup_code:
                        hz = lookup_hz
                if hz is None:
                    raise ValueError
                log.info(
                    "CMD_MEASUREMENT_RATE {} -> set measurement rate to {} "
                    "Hz; RESPONSE_NONSPECIFIC_OK".format(rest, hz))
                self.read_frequency_hz = hz
                self.read_period_s = 1 / hz
                self.reply(RESPONSE_NONSPECIFIC_OK)
            except (TypeError, ValueError):
                log.info("CMD_MEASUREMENT_RATE {} [duff] -> "
                         "RESPONSE_UNKNOWN (?)".format(rest))
                self.reply(RESPONSE_UNKNOWN)

        elif line == CMD_NO_OP:
            log.info("CMD_NO_OP -> clear buffer, no reply")

        elif line == CMD_QUERY_IDENTIFICATION:
            log.info("CMD_QUERY_IDENTIFICATION -> {}".format(
                self.IDENTIFICATION))
            self.reply(self.IDENTIFICATION)

        elif line.startswith(CMD_QUERY_MEASURE):
            rest = line[len(CMD_QUERY_MEASURE):]
            try:
                n = int(rest)
                log.info("CMD_QUERY_MEASURE {n} -> send {n} measurements; no "
                         "other reply".format(n=n))
                self.start_measuring(n)
            except (TypeError, ValueError):
                log.info("CMD_QUERY_MEASURE {} [duff] -> "
                         "RESPONSE_UNKNOWN (?)".format(rest))
                self.reply(RESPONSE_UNKNOWN)

        elif line == CMD_QUERY_OUTPUT_SCALING:
            log.info("CMD_QUERY_OUTPUT_SCALING -> {}".format(
                self.DEFAULT_MAX_VALUE))
            self.reply(str(self.DEFAULT_MAX_VALUE))

        elif line == CMD_QUERY_STATUS:
            log.info("CMD_QUERY_STATUS -> {}".format(self.REPLY_STATUS))
            self.reply(self.REPLY_STATUS)

        elif line == CMD_QUERY_BAUD_RATE:
            result = self.baudrate_string()
            delay_s = 0.5 * BAUDRATE_PAUSE_MS / 1000
            log.info("CMD_QUERY_BAUD_RATE -> pause {} s -> {}".format(
                delay_s, result))
            sleep(delay_s)
            self.reply(result)

        elif line.startswith(CMD_SET_BAUD_RATE):
            # MUST BE AFTER line.startswith(CMD_QUERY_BAUD_RATE),
            # since this is a substring of that
            rest = line[len(CMD_SET_BAUD_RATE):]
            log.info("CMD_SET_BAUD_RATE {} -> RESPONSE_NONSPECIFIC_OK".format(
                rest))
            self.reply(RESPONSE_NONSPECIFIC_OK)

        elif line.startswith(CMD_SET_FILTER):
            rest = line[len(CMD_SET_FILTER):]
            log.info("CMD_SET_FILTER {} -> RESPONSE_NONSPECIFIC_OK".format(
                rest))
            self.reply(RESPONSE_NONSPECIFIC_OK)

        elif line == CMD_STOP_MEASURING:
            log.info("CMD_STOP_MEASURING -> stop measuring, no reply")
            self.stop_measuring()

        elif line == CMD_TARE:
            log.info("CMD_TARE -> RESPONSE_NONSPECIFIC_OK")
            self.reply(RESPONSE_NONSPECIFIC_OK)

        elif line == CMD_WARM_RESTART:
            log.info("CMD_WARM_RESTART -> wait {} s, clear queue, "
                     "no reply".format(self.RESTART_TIME_S))
            self.clear_queue()
            sleep(self.RESTART_TIME_S)

        else:
            log.info("Unknown command: {} -> RESPONSE_UNKNOWN".format(
                repr(line)))
            self.reply(RESPONSE_UNKNOWN)

    def baudrate_string(self) -> str:
        if self.parity == 'E':
            parity_code = 1
        elif self.parity == 'N':
            parity_code = 0
        else:
            parity_code = 2  # no idea
        return "{:05d},{}".format(self.baudrate, parity_code)

    def start_measuring(self, n_measurements) -> None:
        log.info(
            "Starting to pretend to measure {} masses; current read frequency "
            "is {} Hz (period {} s)".format(
                n_measurements,
                self.read_frequency_hz,
                self.read_period_s,
            ))
        for i in range(n_measurements):
            delay_s = 0 if i == 0 else self.read_period_s
            mass_kg = random.gauss(self.current_true_mass_kg, self.MASS_KG_SD)
            balance_perceived_mass_kg = mass_kg - self.tare_mass_kg
            value = int(balance_perceived_mass_kg * self.INTEGER_UNITS_PER_KG)
            value_text = str(value)
            log.info(
                "QUEUING {i}/{n}: True mass {mass} kg - tare {tare} kg = "
                "perceived mass {perc} kg -> value {val}".format(
                    i=i+1,
                    n=n_measurements,
                    mass=mass_kg,
                    tare=self.tare_mass_kg,
                    perc=balance_perceived_mass_kg,
                    val=value))
            om = OutputMessage(delay_s=delay_s, is_measurement=True,
                               text=value_text)
            self.output_queue.append(om)

    def stop_measuring(self) -> None:
        log.info("Stopping mass reading")
        self.output_queue = [x for x in self.output_queue
                             if not x.is_measurement]


# =============================================================================
# Command-line framework
# =============================================================================

def main() -> None:
    # -------------------------------------------------------------------------
    # Arguments
    # -------------------------------------------------------------------------
    parser = argparse.ArgumentParser(
        description="test_mimic_balance -- to test Starfeeder",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
        # formatter_class=lambda prog: argparse.ArgumentDefaultsHelpFormatter(
        #     prog, max_help_position=30, width=120)
        # http://stackoverflow.com/questions/5462873/control-formatting-of-the-argparse-help-argument-list  # noqa
    )
    parser.add_argument(
        '--verbose', '-v', action='store_true', help="Be verbose")
    add_serial_port_args(parser)
    args = parser.parse_args()
    # -------------------------------------------------------------------------
    # Logging
    # -------------------------------------------------------------------------
    loglevel = logging.DEBUG if args.verbose else logging.INFO
    logging.basicConfig(format=LOG_FORMAT, datefmt=LOG_DATEFMT,
                        level=loglevel)
    rootlogger = logging.getLogger()
    rootlogger.setLevel(loglevel)
    configure_logger_for_colour(rootlogger)  # configure root logger
    # -------------------------------------------------------------------------
    # Go
    # -------------------------------------------------------------------------
    log.info("test_mimic_balance")
    log.info("args: {}".format(args))

    mimic = BalanceMimic(args)
    mimic.run()


if __name__ == '__main__':
    main()
